import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.service.roles.agent import RoleAgentController
from abstra_internals.repositories.project.project import ProjectRepository


def get_player_bp(main_controller: MainController) -> flask.Blueprint:
    controller = RoleAgentController(repos=main_controller.repositories)

    bp = flask.Blueprint("role_agent", __name__)

    @bp.post("/")
    def _create_task():
        if not flask.request.json:
            flask.abort(400)
        auth = flask.request.headers.get("authorization")
        if auth is None:
            raise Exception("Authorization header is required")
        payload = flask.request.json
        task_data = payload["task_data"]
        target_stage_id = payload["target_stage_id"]
        controller.create_foreign_task(
            task_type=task_data["type"],
            task_payload=task_data["payload"],
            target_stage_id=target_stage_id,
        )
        return {"success": True}

    @bp.get("/entrypoints")
    def _get_agent_entrypoints():
        project = ProjectRepository.load()
        return [ae.model_dump() for ae in project.get_agent_entrypoints()]

    @bp.post("/connection")
    def _create_connection():
        if not flask.request.json:
            flask.abort(400)
        auth = flask.request.headers.get("authorization")
        if auth is None:
            flask.abort(401)
        proof = auth.replace("Bearer ", "")
        connection = controller.repos.role_agents.create_connection_on_cloud(
            client_proof=proof,
            client_tasks_url=flask.request.json["client_tasks_url"],
            client_stage_id=flask.request.json["client_stage_id"],
            agent_stage_id=flask.request.json["agent_stage_id"],
        )
        return connection.model_dump()

    @bp.patch("/connection")
    def _update_connection():
        if not flask.request.json:
            flask.abort(400)

        auth = flask.request.headers.get("authorization")
        assert auth is not None, "Authorization header is required"

        client_tasks_url = flask.request.json["client_tasks_url"]
        assert client_tasks_url is not None, "client_tasks_url is required"

        connection = controller.repos.role_agents.update_connection_on_cloud(
            connection_token=auth,
            client_tasks_url=client_tasks_url,
        )
        return connection.model_dump()

    @bp.delete("/connection")
    def _delete_connection():
        connection_token = flask.request.headers.get("authorization")
        assert connection_token is not None, "Authorization header is required"

        controller.repos.role_agents.delete_connection_on_cloud(
            connection_token=connection_token
        )
        return {"success": True}

    return bp



from typing import Dict, List, Optional, TypeVar, Union

from abstra_internals.controllers.sdk_ai import Format, Prompt
from abstra_internals.controllers.sdk_context import SDKContextStore

T = TypeVar("T")


def to_list(value: Union[T, List[T], None]) -> List[T]:
    if value is None:
        return []
    if isinstance(value, list):
        return value
    return [value]


def normalize_format(format: Dict[str, object]) -> Dict[str, object]:
    for key, value in format.items():
        if isinstance(value, str):
            format[key] = {"type": value}
        elif isinstance(value, list):
            format[key] = {"enum": value}
        elif value is bool:
            format[key] = {"type": "boolean"}
        elif value is int:
            format[key] = {"type": "integer"}
        elif value is float:
            format[key] = {"type": "number"}
        elif value is str:
            format[key] = {"type": "string"}
        else:
            format[key] = value
    return format


def prompt(
    prompt: Union[Prompt, List[Prompt]],
    instructions: Union[str, List[str]] = [],
    format: Optional[Format] = None,
    temperature: float = 1.0,
):
    if temperature < 0.0 or temperature > 2.0:
        raise ValueError("Temperature must be between 0.0 and 2.0")

    instructions_list = to_list(instructions)
    normalized_format = normalize_format(format) if format else None
    prompt_list = to_list(prompt)

    return SDKContextStore.get_by_thread().ai_sdk.prompt(
        prompt_list, instructions_list, normalized_format, temperature
    )



from multiprocessing.forkserver import set_forkserver_preload

from abstra_internals.controllers import execution_process
from abstra_internals.controllers.execution_consumer import ExecutionConsumer
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import DEFAULT_PORT, RABBITMQ_CONNECTION_URI
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.consumer import RabbitConsumer
from abstra_internals.repositories.factory import get_prodution_app_repositories
from abstra_internals.settings import SettingsController
from abstra_internals.signals import SignalHandlers


def run():
    SignalHandlers.init()
    AbstraLogger.init("cloud")
    SettingsController.set_root_path(".")
    SettingsController.set_server_port(DEFAULT_PORT)

    if not RABBITMQ_CONNECTION_URI:
        raise Exception("RABBITMQ_CONNECTION_URI not found")

    set_forkserver_preload([execution_process.__name__])
    controller = MainController(repositories=get_prodution_app_repositories())
    with RabbitConsumer(RABBITMQ_CONNECTION_URI) as consumer:
        SignalHandlers.register_sigterm_callback(consumer.stop)
        ExecutionConsumer(consumer, controller)


if __name__ == "__main__":
    run()



from abstra_internals.email_templates.html_wrapper import style_wrap
from abstra_internals.email_templates.i18n import get_translation
from abstra_internals.repositories.email import EmailParams
from abstra_internals.repositories.project.project import ProjectRepository

template = """
<p
    style="font-style: normal; font-weight: 700; font-size: 30px; line-height: 36px; color:#181818; margin-bottom:20px; margin-top:30px;">
    {authenticate_with}
</p>
<p style="font-style: normal; font-weight: 400; font-size: 16px; line-height: 24px; color:#252525; margin:0">
    {enter_confirmation_code}
</p>
<p
    style="font-style: normal; font-weight: 700; font-size: 30px; line-height: 36px; color:#181818; margin-bottom:20px; margin-top:30px;">
    {code}
</p>
<div>
    <img style="height: 21px; display:inline; vertical-align:middle;"
        src="https://abstra-cloud-assets.s3.amazonaws.com/bi_hourglass-split+(1).png" />
    <p
        style="font-style: normal; font-weight: 700; font-size: 16px; line-height:21px; color:#707070; margin:0; display:inline; vertical-align:middle;">
        {this_code_expires}
    </p>
</div>
"""


def generate_email(email: str, code: str) -> EmailParams:
    project = ProjectRepository.load()
    translations = get_translation(project.workspace.language or "en")

    content = template.format(
        authenticate_with=translations.confirm_login_at(project.workspace.project_name),
        enter_confirmation_code=translations.enter_verification_code(),
        code=code,
        this_code_expires=translations.this_code_expires(),
    )

    html = style_wrap(content, project.workspace)

    return EmailParams(
        kind="passwordless",
        to=[email],
        subject=translations.confirm_login_at(project.workspace.project_name),
        body=html,
        is_html=True,
        attachments=[],
    )



from typing import List, Optional, Tuple, Union

from abstra_internals.entities.forms.form_state import Button, State
from abstra_internals.entities.forms.template import (
    GeneratorFunc,
    Template,
    TemplateFunc,
)
from abstra_internals.utils.code import has_none_return


class PageStep:
    def __init__(self, func: TemplateFunc):
        self.call = func
        self.computation_step = has_none_return(func)


class GeneratorPageStep:
    def __init__(self, func: GeneratorFunc):
        self.func = func
        self.generator = None
        self.computation_step = False

    def initialize_generator(self, state: State):
        if self.generator is None:
            self.generator = self.func(state)
        return self.generator

    def call(
        self, state: State
    ) -> Union[Template, Tuple[Template, Optional[List[Button]]]]:
        generator = self.initialize_generator(state)

        try:
            return next(generator)
        except StopIteration:
            return []


class EndPageStep(PageStep):
    def __init__(self, func: TemplateFunc):
        super().__init__(func)


Step = Union[PageStep, EndPageStep, GeneratorPageStep]



import base64
import io
import mimetypes
import os
from typing import List, Union

from abstra_internals.email_templates.i18n import get_translation
from abstra_internals.repositories.email import EmailParams
from abstra_internals.repositories.project.project import ProjectRepository

DEFAULT_FILENAME = "attachment.bin"
DEFAULT_APPLICATION_TYPE = "application/octet-stream"


def is_base64(s: str) -> bool:
    try:
        if isinstance(s, str):
            return base64.b64encode(base64.b64decode(s)).decode("utf-8") == s
        return False
    except Exception:
        return False


def convert_to_base64(file: Union[io.IOBase, str]) -> str:
    if isinstance(file, io.IOBase):
        file.seek(0)
        return base64.b64encode(file.read()).decode("utf-8")
    elif isinstance(file, str):
        if is_base64(file):
            return file
        else:
            if os.path.isfile(file):
                with open(file, "rb") as f:
                    return base64.b64encode(f.read()).decode("utf-8")
            else:
                raise ValueError(
                    f"Attachment string is neither a valid base64 nor a valid file path: {file}"
                )


def get_attachment_name(attachment: Union[str, io.IOBase]) -> str:
    if isinstance(attachment, io.FileIO):
        return os.path.basename(attachment.name)
    elif isinstance(attachment, str):
        return os.path.basename(attachment)
    else:
        return DEFAULT_FILENAME


def get_attachment_type(attachment_name: str) -> str:
    mime_type, _ = mimetypes.guess_type(attachment_name)
    return mime_type or DEFAULT_APPLICATION_TYPE


def generate_email(
    to: List[str], message: str, title: str, attachments: List[Union[str, io.IOBase]]
) -> EmailParams:
    project = ProjectRepository.load()
    translations = get_translation(project.workspace.language or "en")

    subject = translations.message_from(project.workspace.project_name)
    if title:
        subject += f": {title}"

    parsed_attachments = []
    for attachment in attachments:
        attachment_name = get_attachment_name(attachment)
        attachment_type = get_attachment_type(attachment_name)
        attachment_content = convert_to_base64(attachment)
        parsed_attachments.append(
            {
                "filename": attachment_name,
                "content": attachment_content,
                "type": attachment_type,
            }
        )
    return EmailParams(
        kind="message",
        to=to,
        subject=subject,
        body=message,
        attachments=parsed_attachments,
        is_html=False,
    )



from typing import Iterator, List, Optional

from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.controllers.sdk_tasks import Task
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.tasks import TaskPayload
from abstra_internals.utils.json import to_json_serializable

BATCH_SIZE = 10


def get_tasks(
    limit: Optional[int] = None, offset=0, where: Optional[dict] = None
) -> List[Task]:
    context = SDKContextStore.get_by_thread()
    if where is None:
        where = {}

    if limit is None or limit < 1:
        limit = 100
    else:
        if limit > 10000:
            limit = 10000
            AbstraLogger.warning(
                "get_tasks limit is too high, capping it to 10000. To iterate over all tasks, use iter_tasks"
            )

    return context.task_sdk.get_stage_pending_tasks(limit, offset, where)


def send_task(type: str, payload: TaskPayload, show_warning=True) -> None:
    context = SDKContextStore.get_by_thread()
    if not isinstance(payload, dict):
        raise Exception("Task payload must be a dictionary")
    serializable_payload = {k: to_json_serializable(v) for k, v in payload.items()}
    context.task_sdk.send_task(type, serializable_payload)


def get_trigger_task() -> Task:
    context = SDKContextStore.get_by_thread()
    return context.task_sdk.get_trigger_task()


def iter_tasks(where: Optional[dict] = None) -> Iterator[Task]:
    if where is None:
        where = {}
    offset = 0
    while True:
        tasks = get_tasks(limit=BATCH_SIZE, offset=offset, where=where)
        if not tasks:
            break
        yield from tasks
        offset += BATCH_SIZE


def get_sent_tasks(
    limit: Optional[int] = None, offset=0, where: Optional[dict] = None
) -> List[Task]:
    context = SDKContextStore.get_by_thread()
    if limit is None and where is None:
        limit = 100
    if where is None:
        where = {}
    return context.task_sdk.get_stage_sent_tasks(limit, offset, where)



from abc import abstractmethod
from typing import Dict, List, Optional

import requests

from abstra_internals.cloud_api import get_project_info
from abstra_internals.credentials import resolve_headers
from abstra_internals.entities.agents import (
    AgentEntrypoint,
    AgentModel,
    ConnectionModel,
    SignProofResponse,
)
from abstra_internals.environment import PROJECT_ID, SIDECAR_HEADERS
from abstra_internals.repositories.services.roles.common import RoleCommonRepository
from abstra_internals.settings import Settings


class RoleClientRepository(RoleCommonRepository):
    _cached_agents: Optional[Dict[str, AgentModel]] = None
    _cached_connections: Optional[List[ConnectionModel]] = None
    base_url: str

    def get_connections(self) -> List[ConnectionModel]:
        return self._cached_connections or []

    @property
    def role(self):
        return "client"

    @abstractmethod
    def get_headers(self):
        raise NotImplementedError

    @abstractmethod
    def get_current_project_id(self):
        raise NotImplementedError

    def sign_proof(self, agent_project_id: str):
        client_project_id = self.get_current_project_id()
        headers = self.get_headers()
        assert headers is not None, "You should be logged in to sign a proof"
        url = f"{self.base_url}/sign-proof"
        response = requests.post(
            url,
            headers=headers,
            json=dict(
                clientProjectId=client_project_id,
                agentProjectId=agent_project_id,
            ),
        )
        if not response.ok:
            raise Exception(response.text)
        return SignProofResponse.model_validate_json(response.text)

    def get_or_create_connection_to_agent(
        self, agent_project_id: str, client_stage_id: str, agent_stage_id: str
    ):
        for connection in self.get_connections():
            if connection.agent_project_id != agent_project_id:
                continue
            if connection.client_stage_id != client_stage_id:
                continue
            if connection.agent_stage_id != agent_stage_id:
                continue
            return connection
        return self.connect_to_agent(agent_project_id, client_stage_id, agent_stage_id)

    def connect_to_agent(
        self, agent_project_id: str, client_stage_id: str, agent_stage_id: str
    ) -> ConnectionModel:
        """
        First step of the protocol. It connects to another abstra project known as an agent.
        """
        url = self.get_agents([agent_project_id])[0].tasks_url + "/agent/connection"
        public_url = Settings.public_url
        assert (
            public_url is not None
        ), "You should be connected to a tunnel to connect to an agent"
        proof = self.sign_proof(agent_project_id=agent_project_id).signedProof
        response = requests.post(
            url,
            headers={
                "Authorization": f"Bearer {proof}",
            },
            json=dict(
                client_tasks_url=public_url + "/_tasks/client",
                agent_project_id=agent_project_id,
                client_proof=proof,
                agent_stage_id=agent_stage_id,
                client_stage_id=client_stage_id,
            ),
        )

        response.raise_for_status()
        connection = ConnectionModel.model_validate(response.json())
        self._cached_connections = self.get_connections() + [connection]
        return connection

    def update_connection(
        self, connection: ConnectionModel, client_tasks_url: str
    ) -> ConnectionModel:
        agent = self.get_agent(connection.agent_project_id)

        url = agent.tasks_url + "/agent/connection"
        response = requests.patch(
            url,
            headers={
                "Authorization": connection.token,
            },
            json=dict(
                client_tasks_url=client_tasks_url,
            ),
        )
        response.raise_for_status()
        connection = ConnectionModel.model_validate(response.json())
        self._cached_connections = self.get_connections() + [connection]
        return connection

    def delete_connection(self, connection: ConnectionModel):
        agent = self.get_agent(connection.agent_project_id)

        url = agent.tasks_url + "/agent/connection"
        response = requests.delete(
            url,
            headers={
                "Authorization": connection.token,
            },
        )
        response.raise_for_status()
        self._cached_connections = [
            c for c in self.get_connections() if c.token != connection.token
        ]

    def get_agents(self, ids: List[str]) -> List[AgentModel]:
        if self._cached_agents is not None and all(
            id in self._cached_agents for id in ids
        ):
            return [self._cached_agents[id] for id in ids]
        else:
            headers = self.get_headers()
            assert headers is not None, "You should be logged in to get agents"
            url = f"{self.base_url}/agents"
            response = requests.get(url, headers=headers, params={"ids": ids})
            if not response.ok:
                raise Exception(response.text)
            agents = [AgentModel.model_validate(agent) for agent in response.json()]
            self._cached_agents = {agent.project_id: agent for agent in agents}
            return agents

    def get_agent(self, id: str):
        return self.get_agents([id])[0]

    def get_entrypoints(self, project_id: str) -> List[AgentEntrypoint]:
        agent = self.get_agent(project_id)
        url = agent.tasks_url + "/agent/entrypoints"
        res = requests.get(url)
        res.raise_for_status()
        return [AgentEntrypoint.model_validate(e) for e in res.json()]


class ProductionRoleClientRepository(RoleClientRepository):
    def get_headers(self):
        return SIDECAR_HEADERS

    def get_current_project_id(self):
        return PROJECT_ID


class LocalRoleClientRepository(RoleClientRepository):
    def get_headers(self):
        return resolve_headers()

    def get_current_project_id(self):
        headers = resolve_headers()
        assert headers is not None, "You should be logged in to get the project id"
        return get_project_info(headers)["id"]



import pkg_resources
from pkg_resources import get_distribution as gd

from abstra_internals.linter.rules.psycopg2 import (
    Psycopg2FoundWithoutBinary,
    Psycopg2MustBeBinary,
)
from tests.fixtures import BaseTest


def mock_get_distribution(name):
    if name == "abstra":
        return pkg_resources.Distribution(project_name="abstra", version="1.0.0")
    else:
        return gd(name)


class MissingAbstraInRequirementsTest(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.old_get_distribution = pkg_resources.get_distribution
        pkg_resources.get_distribution = mock_get_distribution

    def tearDown(self) -> None:
        pkg_resources.get_distribution = self.old_get_distribution
        super().tearDown()

    def test_psycopg2_valid(self):
        rule = Psycopg2MustBeBinary()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("psycopg2-binary")
        self.assertEqual(len(rule.find_issues()), 0)

    def test_psycopg2_invalid(self):
        rule = Psycopg2MustBeBinary()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("psycopg2")
        issues = rule.find_issues()
        self.assertEqual(len(issues), 1)
        self.assertIsInstance(issues[0], Psycopg2FoundWithoutBinary)



import inspect
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Optional, Tuple

import flask
import requests

from abstra_internals.credentials import get_credentials
from abstra_internals.environment import CLOUD_API_CLI_URL, SIDECAR_HEADERS, SIDECAR_URL
from abstra_internals.jwt_auth import USER_AUTH_HEADER_KEY, UserClaims
from abstra_internals.threaded import threaded
from abstra_internals.utils import get_local_python_version, get_local_user_id
from abstra_internals.utils.env import is_dev_env, is_test_env
from abstra_internals.utils.packages import get_local_package_version


# EDITOR
@threaded
def send_editor_usage(payload: Dict):
    if is_test_env() or is_dev_env():
        return

    data = {
        "payload": payload,
        "userId": get_local_user_id(),
        "pythonVersion": get_local_python_version(),
        "abstraVersion": str(get_local_package_version()),
    }

    headers = {"apiKey": get_credentials()}
    api_url = f"{CLOUD_API_CLI_URL}/editor/usage"
    requests.post(api_url, json=data, headers=headers)


def editor_usage(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Tuple[Any], **kwargs: Any) -> Any:
        arg_names = inspect.getfullargspec(func).args
        arg_values = dict(zip(arg_names, args))

        send_editor_usage({**arg_values, **kwargs, **{"event": func.__name__}})
        return func(*args, **kwargs)

    return wrapper


def editor_manual_usage(*, event: str, payload: Dict):
    send_editor_usage({**payload, "event": event})


# PLAYER
@threaded
def send_player_usage(*, event: str, payload: Dict, auth: Optional[str] = None):
    if SIDECAR_URL is None:
        return

    body = dict(event=event, payload=payload)
    if auth is not None:
        if claims := UserClaims.from_jwt(auth.split(" ")[1]):
            body["email"] = claims.email

    requests.post(f"{SIDECAR_URL}/usage/event", headers=SIDECAR_HEADERS, json=body)


def player_usage(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Tuple[Any], **kwargs: Any) -> Any:
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        arg_names = inspect.getfullargspec(func).args
        arg_values = dict(zip(arg_names, args))

        send_player_usage(
            auth=auth, event=func.__name__, payload={**arg_values, **kwargs}
        )
        return func(*args, **kwargs)

    return wrapper


# Executions
def _send_execution_usage(file: Path, status: str, exception: Optional[Exception]):
    if SIDECAR_URL or is_test_env() or is_dev_env():
        return

    try:
        editor_manual_usage(
            event="execution_exception",
            payload=dict(file=str(file), status=status, exception=str(exception)),
        )
    except Exception:
        pass


Result = Tuple[Literal["finished", "abandoned", "failed"], Optional[Exception]]


def execution_usage(func: Callable[[Path], Result]) -> Callable[[Path], Result]:
    @wraps(func)
    def wrapper(file: Path) -> Result:
        status, exception = func(file)
        _send_execution_usage(file, status, exception)
        return status, exception

    return wrapper



import datetime
from dataclasses import dataclass
from typing import Any, List
from uuid import uuid4

from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.entities.execution_context import (
    FormContext,
    HookContext,
    Request,
    ScriptContext,
)
from abstra_internals.interface.sdk import user_exceptions
from abstra_internals.repositories.tasks import TaskDTO, TaskEventDetails, TaskPayload
from abstra_internals.utils.datetime import to_utc_iso_string


@dataclass
class MockedTask:
    type: str
    payload: Any


def mock_form_answers(test_answers: List[Any]):
    context = SDKContextStore.get_by_thread()
    if not isinstance(context.task_sdk.execution.context, FormContext):
        raise user_exceptions.IncompatibleTestingFormsSDK()
    context.form_sdk.client.context.mock_execution.test_answers = test_answers


def mock_trigger_task(type: str, payload: TaskPayload):
    context = SDKContextStore.get_by_thread()
    if not isinstance(context.task_sdk.execution.context, ScriptContext):
        raise user_exceptions.IncompatibleTestingOnTasksSDK()
    mocked_task_id = "test-" + str(uuid4())
    context.task_sdk.execution.context.mock_execution.test_trigger_task = TaskDTO(
        id=mocked_task_id,
        type=type,
        payload=payload,
        target_stage_id="test",
        status="pending",
        created=TaskEventDetails(
            at=to_utc_iso_string(datetime.datetime.now(datetime.timezone.utc)),
            by_execution_id=context.task_sdk.execution.id,
            by_stage_id=context.task_sdk.execution.stage_id,
        ),
        locked=None,
        completed=None,
    )


def mock_hook_request(
    method: str,
    body: str = "",
    query_params: dict = {},
    headers: dict = {},
):
    context = SDKContextStore.get_by_thread()
    if not isinstance(context.task_sdk.execution.context, HookContext):
        raise user_exceptions.IncompatibleTestingHooksSDK()
    mocked_request = Request(
        method=method, body=body, query_params=query_params, headers=headers
    )
    context.task_sdk.execution.context.mock_execution.test_request = mocked_request


def mock_pending_tasks(tasks: List[dict]):
    context = SDKContextStore.get_by_thread()
    context.task_sdk.execution.context.mock_execution.test_pending_tasks = [
        TaskDTO(
            id="test-" + str(uuid4()),
            type=t["type"],
            payload=t["payload"],
            target_stage_id="test",
            status="pending",
            created=TaskEventDetails(
                at=to_utc_iso_string(datetime.datetime.now(datetime.timezone.utc)),
                by_execution_id=context.task_sdk.execution.id,
                by_stage_id=context.task_sdk.execution.stage_id,
            ),
            locked=None,
            completed=None,
        )
        for t in tasks
    ]



import time
from threading import Thread
from typing import List, Optional

from abstra_internals.controllers.common.task_executors import TaskExecutor
from abstra_internals.controllers.service.roles.common import RoleCommonController
from abstra_internals.entities.execution import Execution
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.settings import Settings


class RoleClientController(RoleCommonController):
    def entrypoints(self, project_id: str) -> List[dict]:
        return [
            e.to_camel() for e in self.repos.role_clients.get_entrypoints(project_id)
        ]

    def create_foreign_task(
        self,
        task_type: str,
        task_payload: dict,
        target_stage_id: str,
        execution: Optional[Execution] = None,
    ):
        executor = TaskExecutor(self.repos)
        project = ProjectRepository.load()
        executor.send_task(
            type=task_type,
            current_stage=project.get_stage_raises(target_stage_id),
            payload=task_payload,
            execution=execution,
        )

    def sync_connection_pool(self):
        if Settings.public_url is None:
            return

        project = ProjectRepository.load()
        connections = self.repos.role_clients.get_connections()

        for connection in connections:
            for agent in project.agents:
                if agent.id == connection.client_stage_id:
                    break
            else:
                self.repos.role_clients.delete_connection(connection)

        for agent in project.agents:
            if not agent.client_stage_id or not agent.project_id:
                continue

            for connection in connections:
                if (
                    connection.agent_stage_id == agent.client_stage_id
                    and connection.agent_project_id == agent.project_id
                    and connection.client_stage_id == agent.id
                ):
                    self.repos.role_clients.update_connection(
                        connection=connection,
                        client_tasks_url=Settings.public_url + "/_tasks",
                    )
                    break

            else:
                self.repos.role_clients.connect_to_agent(
                    agent_stage_id=agent.client_stage_id,
                    agent_project_id=agent.project_id,
                    client_stage_id=agent.id,
                )

    def loop_sync_connection_pool(self):
        def loop():
            while True:
                if Settings.has_public_url():
                    try:
                        self.sync_connection_pool()
                        time.sleep(60 * 5)
                    except Exception as e:
                        print(e)
                        pass
                else:
                    time.sleep(10)

        Thread(target=loop, daemon=True).start()



from abstra_internals.repositories.project.project import AgentStage, ProjectRepository


def add_agent(agent_project_id: str, agent_title: str):
    stage = AgentStage.create(
        title=agent_title,
        project_id=agent_project_id,
    )
    project = ProjectRepository.load()
    project.add_stage(stage)
    ProjectRepository.save(project)



import copy
from typing import Callable, Dict, Generator, List, Optional, Tuple, TypedDict, Union

from abstra_internals.entities.forms.form_state import Button, State
from abstra_internals.widgets.widget_base import Input, Widget

Template = List[Widget]
TemplateFunc = Callable[
    [State], Union[Template, Tuple[Template, Optional[List[Button]]]]
]
RawTemplate = Union[Template, TemplateFunc]
GeneratorFunc = Callable[
    [State],
    Generator[Union[Template, Tuple[Template, Optional[List[Button]]]], None, None],
]


LEGACY_CTX = {}


class RenderOutput(TypedDict):
    widgets: List[Dict]
    has_errors: bool


class TemplateRenderer:
    def __init__(self, template: Template) -> None:
        self._raw_template = template

    @property
    def template(self) -> Template:
        # Avoids using mutable objects as widgets currently have a internal state
        return copy.deepcopy(self._raw_template)

    def parse_state(self, raw_state: Dict) -> State:
        parsed = {}
        for widget in self.template:
            if isinstance(widget, Input):
                parsed[widget.key] = widget.parse_value(raw_state.get(widget.key))
        return State(parsed)

    def render(self, state: State) -> RenderOutput:
        rendered_widgets = []
        has_errors = False
        for widget in self.template:
            if isinstance(widget, Input):
                value = state.get(widget.key)
                if value is not None:
                    widget.set_value(value)
                if state.touched(widget.key):
                    widget.set_errors()
                if widget.has_errors() > 0:
                    has_errors = True

            rendered = widget.render(LEGACY_CTX)
            rendered_widgets.append(rendered)

        return RenderOutput(widgets=rendered_widgets, has_errors=has_errors)



from datetime import datetime
from typing import List, Optional, Tuple, Union
from abstra_internals.widgets.response_types import AppointmentSlot
from abstra_internals.widgets.widget_base import Input


class AppointmentInput(Input):
    type = 'appointment-input'
    empty_value = None
    value: Optional[AppointmentSlot]
    slots: List[AppointmentSlot]

    def __init__(self, key: str, label: str, **kwargs):
        initial_error = kwargs.get('errors', [])
        if not isinstance(initial_error, list):
            initial_error = [initial_error]
        super().__init__(key, initial_error)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)
        self.value = self.load_initial_value(props.get('initial_value',
            self.empty_value))
        self.slots = AppointmentSlot.from_list(props.get('slots', []))

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'hint': self.hint, 'value': self.serialize_value(), 'required':
            self.required, 'fullWidth': self.full_width, 'disabled': self.
            disabled, 'slots': [s.to_dict() for s in self.slots], 'errors':
            self.errors}

    def serialize_value(self) ->Optional[int]:
        return self.slots.index(self.value) if self.value is not None else None

    def parse_value(self, value: Optional[int]) ->Optional[AppointmentSlot]:
        if value is None:
            return None
        return self.slots[value]

    def load_initial_value(self, value: Union[AppointmentSlot, dict, Tuple[
        datetime, datetime]]) ->Optional[AppointmentSlot]:
        if isinstance(value, AppointmentSlot):
            return value
        elif isinstance(value, dict):
            return AppointmentSlot.from_dict(value)
        elif isinstance(value, tuple):
            return AppointmentSlot(*value)
        return None



from concurrent.futures import ThreadPoolExecutor
from uuid import uuid4

from abstra_internals.controllers.execution_process import ExecutionProcess
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import QUEUE_CONCURRENCY
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.consumer import Consumer, QueueMessage
from abstra_internals.settings import Settings


class StageNotFound(Exception):
    pass


class NonCleanExit(Exception):
    pass


# should be smaller than the RabbitMQ consumer timeout (24h now)
PROCESS_TIMEOUT = 6 * 60 * 60


def PreExecController(
    *,
    controller: MainController,
    consumer: Consumer,
    msg: QueueMessage,
    arbiter_uuid: str,
):
    worker_uuid = str(uuid4())
    head_id = worker_uuid.split("-")[0]

    try:
        mp_context = controller.repositories.mp_context.get_context()
        stage = controller.get_action(msg.preexecution.stage_id)
        if stage is None:
            raise StageNotFound(msg.preexecution.stage_id)

        p = mp_context.Process(
            target=ExecutionProcess,
            kwargs=dict(
                stage=stage,
                controller=controller,
                worker_uuid=worker_uuid,
                arbiter_uuid=arbiter_uuid,
                root_path=Settings.root_path,
                server_port=Settings.server_port,
                request=msg.preexecution.context,
                environment=AbstraLogger.environment,
            ),
            name=f"Worker-{head_id}",
        )

        p.start()
        p.join(timeout=PROCESS_TIMEOUT)
        consumer.done_callback(msg)

        if p.is_alive():
            p.terminate()
            raise NonCleanExit(f"Worker took too long to complete ({PROCESS_TIMEOUT})")

        if p.exitcode != 0:
            err_msg = f"Worker exited with status ({p.exitcode})"
            if p.exitcode == -9:
                err_msg += ": Server reached its memory limit"

            raise NonCleanExit(err_msg)
    except Exception as e:
        AbstraLogger.error(f"[{head_id}] PreExecController ERROR: aborting consumer")
        AbstraLogger.capture_exception(e)
        controller.child_exit(
            app_id=arbiter_uuid, worker_id=worker_uuid, err_msg=f"[ABORTED] {e.args}"
        )
        consumer.stop()


class Arbiter:
    def __init__(self, controller: MainController):
        self.controller = controller
        self.uuid = str(uuid4())

    @property
    def head_id(self) -> str:
        return self.uuid.split("-")[0]

    def __enter__(self):
        AbstraLogger.debug(f"[{self.head_id}] ARBITER INIT")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        AbstraLogger.debug(f"[{self.head_id}] ARBITER EXIT")
        self.controller.self_exit(app_id=self.uuid, err_msg="[ABORTED] Worker exited")


def ExecutionConsumer(consumer: Consumer, controller: MainController):
    with Arbiter(controller) as arbiter:
        with ThreadPoolExecutor(max_workers=QUEUE_CONCURRENCY) as executor:
            try:
                for msg in consumer.iter():
                    executor.submit(
                        PreExecController,
                        arbiter_uuid=arbiter.uuid,
                        controller=controller,
                        consumer=consumer,
                        msg=msg,
                    )
            except Exception as e:
                AbstraLogger.capture_exception(e)



import abc
from dataclasses import dataclass

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import SIDECAR_HEADERS


@dataclass
class AccessTokenDTO:
    connectionName: str
    connectorType: str
    projectId: str
    token: str
    expiresAt: str


class ConnectorsRepository(abc.ABC):
    url: str

    def __init__(self, base_url: str) -> None:
        self.url = f"{base_url}/connectors"

    def get_access_token_url(self, connection_name: str) -> str:
        return f"{self.url}/connection/{connection_name}/access-token"

    def get_headers(self) -> dict:
        raise NotImplementedError()

    def get_access_token(self, connection_name: str) -> AccessTokenDTO:
        response = requests.get(
            self.get_access_token_url(connection_name), headers=self.get_headers()
        )
        response.raise_for_status()
        return AccessTokenDTO(**response.json())


class ProductionConnectorsRepository(ConnectorsRepository):
    def get_headers(self) -> dict:
        return SIDECAR_HEADERS


class LocalConnectorsRepository(ConnectorsRepository):
    def get_headers(self) -> dict:
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to execute a table query")
        return headers



from dataclasses import dataclass
from typing import Dict


class AutoNext:
    pass


@dataclass
class Button:
    label: str


class NextButton(Button):
    def __init__(self):
        super().__init__("i18n_next_action")


class BackButton(Button):
    def __init__(self):
        super().__init__("i18n_back_action")


class State(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _touched(self) -> Dict[str, bool]:
        if "__touched" not in self:
            self["__touched"] = {}
        return self["__touched"]

    @property
    def value(self) -> Dict:
        return self.clean_dict_from_touched(self)

    def clean_dict_from_touched(self, obj: Dict) -> Dict:
        if isinstance(obj, dict):
            return {
                k: self.clean_dict_from_touched(v)
                for k, v in obj.items()
                if k != "__touched"
            }
        elif isinstance(obj, list):
            return [self.clean_dict_from_touched(v) for v in obj]
        else:
            return obj

    def touch_all(self):
        self._touched.update({key: True for key in self.keys()})
        for value in self.values():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, State):
                        item.touch_all()

    def touched(self, key: str) -> bool:
        return key in self._touched

    def init_keys(self, other: Dict):
        for key, value in other.items():
            # print(f"Warning: key {key} not found in state")

            if key not in self:
                # print(f"Warning: key {key} not found in state")
                self[key] = value

    def merge(self, other: Dict):
        for key, value in other.items():
            self._touched[key] = True
            if isinstance(value, list):
                for i, item in enumerate(value):
                    if i < len(self[key]):
                        self[key][i].merge(item)
                    else:
                        self[key].append(item)
            elif value != self.get(key):
                self[key] = value



import types
from typing import Dict, Generator, List, Optional, Union

from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.controllers.sdk_forms import FormSDKController
from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.steps import (
    EndPageStep,
    GeneratorPageStep,
    PageStep,
    Step,
)
from abstra_internals.entities.forms.template import (
    GeneratorFunc,
    Template,
    TemplateFunc,
)
from abstra_internals.interface.sdk.forms.exceptions import InvalidRunInputError
from abstra_internals.widgets.widget_base import Input, Widget

Runnable = Union[TemplateFunc, Template, GeneratorFunc]


class Form:
    controller: FormSDKController
    state: State
    hide_steps: bool

    def _create_step(self, runnable: Runnable) -> Step:
        if isinstance(runnable, EndPageStep):
            return runnable
        elif isinstance(runnable, types.FunctionType):
            if isinstance(runnable(State()), Generator):
                return GeneratorPageStep(runnable)
            return PageStep(runnable)
        elif isinstance(runnable, list):
            return PageStep(lambda _: (runnable, None))

        raise InvalidRunInputError(str(type(runnable)))

    def __init__(
        self,
        runnables: List[Runnable],
        initial_state: Optional[Dict] = None,
        hide_steps: bool = False,
    ):
        self.steps = [self._create_step(runnable) for runnable in runnables]
        self.controller = SDKContextStore.get_by_thread().form_sdk
        self.state = State(initial_state or {})
        self.hide_steps = hide_steps
        if len(list(filter(lambda s: not s.computation_step, self.steps))) == 1:
            self.hide_steps = True

    def run(self) -> Dict:
        return self.controller.run_form(self.steps, self.state, self.hide_steps)


# TODO: Do we want multiple return types here?
def run(
    runnables: Union[List[Runnable], Widget],
    state: Optional[Dict] = None,
    hide_steps: bool = False,
) -> Union[State, None, object]:
    if isinstance(runnables, Widget):
        state = Form([[runnables]], state, hide_steps).run()
        if isinstance(runnables, Input):
            return state[runnables.key]
        return None

    return Form(runnables, state, hide_steps).run()



from typing import Dict, List, Optional, TypedDict

from abstra_internals.controllers.main import UnknownNodeTypeError
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import (
    AgentStage,
    ClientStage,
    FormStage,
    HookStage,
    JobStage,
    NotificationTrigger,
    Project,
    ProjectRepository,
    ScriptStage,
    Stage,
    WorkflowTransition,
)


class StageDTO(TypedDict):
    id: str
    type: str
    title: str
    position: Dict[str, float]
    props: Dict[str, Optional[str]]


class WorkflowController:
    def __init__(self, repos: Repositories):
        self.repos = repos

    # workflow visual editor
    def get_workflow(self):
        project = ProjectRepository.load()
        return self._make_workflow_dto(project)

    def build_adjacency_list(
        self,
    ) -> Dict[str, List[str]]:
        workflow = self.get_workflow()
        transitions = workflow["transitions"]
        adj = {}
        for t in transitions:
            if t["sourceStageId"] not in adj:
                adj[t["sourceStageId"]] = []
            adj[t["sourceStageId"]].append(t["targetStageId"])
        return adj

    def get_initial_stages(
        self,
    ):
        workflow = self.get_workflow()
        stages = workflow["stages"]
        target_stage_ids = set([t["targetStageId"] for t in workflow["transitions"]])
        return [s for s in stages if s["id"] not in target_stage_ids]

    def get_path(self, n: int):
        adj = self.build_adjacency_list()
        initial_stages = self.get_initial_stages()
        longest_path = []

        def dfs(stage_id: str, current_path: List[str]):
            if stage_id in current_path:
                return
            current_path = current_path.copy()
            current_path.append(stage_id)
            if len(current_path) > len(longest_path):
                longest_path.clear()
                longest_path.extend(current_path)
            for next_stage_id in adj.get(stage_id, []):
                dfs(next_stage_id, current_path)

        for stage in initial_stages:
            dfs(stage["id"], [])

        return longest_path[:n]

    def make_stage_dto(self, stage: Stage) -> StageDTO:
        filename = None
        props = {}
        if isinstance(stage, (HookStage, ScriptStage, FormStage, JobStage)):
            filename = stage.file
            props["filename"] = filename
        path = None
        if isinstance(stage, (FormStage, HookStage)):
            path = stage.path
            props["path"] = path
        project_id = None
        if isinstance(stage, AgentStage):
            project_id = stage.project_id
            props["projectId"] = project_id
            props["clientStageId"] = stage.client_stage_id
        return StageDTO(
            id=stage.id,
            type=stage.type_name + "s",
            title=stage.title,
            position=dict(
                x=stage.workflow_position[0],
                y=stage.workflow_position[1],
            ),
            props=props,
        )

    def _make_workflow_dto(self, project: Project):
        stages = []
        for stage in project.workflow_stages:
            stage_dto = self.make_stage_dto(stage)

            if (
                isinstance(stage, FormStage)
                or isinstance(stage, ScriptStage)
                or isinstance(stage, HookStage)
                or isinstance(stage, JobStage)
            ):
                stage_dto["props"]["filename"] = stage.file

            if isinstance(stage, FormStage) or isinstance(stage, HookStage):
                stage_dto["props"]["path"] = stage.path

            if isinstance(stage, AgentStage):
                stage_dto["props"]["projectId"] = stage.project_id
                stage_dto["props"]["clientStageId"] = stage.client_stage_id

            stages.append(stage_dto)

        transitions = []
        for stage in project.workflow_stages:
            for transition in stage.workflow_transitions:
                props: dict = {"taskType": None}

                if transition.task_type is not None:
                    props["taskType"] = transition.task_type

                transitions.append(
                    dict(
                        id=transition.id,
                        sourceStageId=stage.id,
                        targetStageId=transition.target_id,
                        type=transition.type,
                        props=props,
                    )
                )

        return dict(
            stages=stages,
            transitions=transitions,
        )

    def update_workflow(self, workflow_state_dto: Dict):
        project = ProjectRepository.load()

        for stage_dto in workflow_state_dto["stages"]:
            stage = project.get_stage(stage_dto["id"])
            if stage is not None:
                stage.workflow_position = (
                    stage_dto["position"]["x"],
                    stage_dto["position"]["y"],
                )
                stage.title = stage_dto["title"]
                if (
                    isinstance(stage, FormStage)
                    or isinstance(stage, ScriptStage)
                    or isinstance(stage, HookStage)
                    or isinstance(stage, JobStage)
                ):
                    stage.file = f"{stage_dto['props']['filename']}"
                if isinstance(stage, FormStage) or isinstance(stage, HookStage):
                    stage.path = stage_dto["props"]["path"]
                if isinstance(stage, AgentStage):
                    stage.project_id = stage_dto["props"].get("projectId", "")
                    stage.client_stage_id = stage_dto["props"].get(
                        "clientStageId", None
                    )
            elif stage_dto["type"] == "forms":
                stage = FormStage(
                    id=stage_dto["id"],
                    path=stage_dto["props"]["path"],
                    file=f"{stage_dto['props']['filename']}",
                    title=stage_dto["title"],
                    workflow_transitions=[],
                    is_initial=False,
                    workflow_position=(
                        stage_dto["position"]["x"],
                        stage_dto["position"]["y"],
                    ),
                    notification_trigger=NotificationTrigger(
                        variable_name="assignee_emails",
                        enabled=False,
                    ),
                )
                project.forms.append(stage)
            elif stage_dto["type"] == "scripts":
                stage = ScriptStage(
                    file=f"{stage_dto['props']['filename']}",
                    is_initial=False,
                    id=stage_dto["id"],
                    title=stage_dto["title"],
                    workflow_position=(
                        stage_dto["position"]["x"],
                        stage_dto["position"]["y"],
                    ),
                    workflow_transitions=[],
                )
                project.scripts.append(stage)
            elif stage_dto["type"] == "hooks":
                stage = HookStage(
                    id=stage_dto["id"],
                    enabled=True,
                    file=f"{stage_dto['props']['filename']}",
                    is_initial=False,
                    path=stage_dto["props"]["path"],
                    title=stage_dto["title"],
                    workflow_position=(
                        stage_dto["position"]["x"],
                        stage_dto["position"]["y"],
                    ),
                    workflow_transitions=[],
                )
                project.hooks.append(stage)
            elif stage_dto["type"] == "jobs":
                stage = JobStage(
                    file=f"{stage_dto['props']['filename']}",
                    id=stage_dto["id"],
                    schedule="0 * * * *",
                    title=stage_dto["title"],
                    workflow_position=(
                        stage_dto["position"]["x"],
                        stage_dto["position"]["y"],
                    ),
                    workflow_transitions=[],
                )
                project.jobs.append(stage)
            elif stage_dto["type"] == "agents":
                stage = AgentStage(
                    id=stage_dto["id"],
                    title=stage_dto["title"],
                    project_id=stage_dto["props"].get("projectId", ""),
                    workflow_position=(
                        stage_dto["position"]["x"],
                        stage_dto["position"]["y"],
                    ),
                    workflow_transitions=[],
                    client_stage_id=stage_dto["props"].get("clientStageId", None),
                )
                project.agents.append(stage)
            elif stage_dto["type"] == "clients":
                stage = ClientStage(
                    id=stage_dto["id"],
                    title=stage_dto["title"],
                    workflow_position=(
                        stage_dto["position"]["x"],
                        stage_dto["position"]["y"],
                    ),
                    workflow_transitions=[],
                )
                project.clients.append(stage)
            else:
                raise UnknownNodeTypeError(stage_dto["type"])

        # Delete removed stages
        for stage in project.workflow_stages:
            if stage.id not in [s["id"] for s in workflow_state_dto["stages"]]:
                project.delete_stage(stage.id)

        for stage in project.workflow_stages:
            stage.workflow_transitions = []
            for transition in workflow_state_dto["transitions"]:
                props = transition.get("props", {})
                task_type = props.get("taskType", None)
                if transition["sourceStageId"] == stage.id:
                    target = project.get_stage(transition["targetStageId"])
                    assert target is not None
                    stage.workflow_transitions.append(
                        WorkflowTransition(
                            id=transition["id"],
                            type=transition["type"],
                            target_id=transition["targetStageId"],
                            target_type=target.type_name + "s",
                            task_type=task_type,
                        )
                    )

        ProjectRepository.save(project)

        return self._make_workflow_dto(project)



import json
from threading import Thread
from time import sleep
from typing import Any, Optional

import requests
import simple_websocket
from colorama import Fore
from pydantic import BaseModel

from abstra_internals.contracts_generated import (
    CloudApiCliApiKeyInfoResponse,
    CloudApiCliBuildCreateResponse,
)
from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import (
    CLOUD_API_CLI_URL,
    CLOUD_API_ENDPOINT,
    HOST,
    IS_PRODUCTION,
    PROJECT_ID,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.settings import Settings


def create_build(headers: dict) -> CloudApiCliBuildCreateResponse:
    url = f"{CLOUD_API_CLI_URL}/builds"
    r = requests.post(url=url, headers=headers)
    if not r.ok:
        raise Exception(f"Failed to create build: {r.text}")

    data = r.json()
    return CloudApiCliBuildCreateResponse.from_dict(data)


def update_build(build_id: str, headers: dict, error: Optional[str] = None):
    url = f"{CLOUD_API_CLI_URL}/builds/{build_id}"
    requests.patch(url=url, headers=headers, json=dict(error=error)).raise_for_status()


def get_api_key_info(headers: dict) -> dict:
    url = f"{CLOUD_API_CLI_URL}/api-keys/info"
    try:
        response = requests.get(url, headers=headers)
    except Exception as e:
        AbstraLogger.capture_exception(e)
        return {"logged": False, "reason": "CONNECTION_ERROR"}

    if response.ok:
        response_data = CloudApiCliApiKeyInfoResponse.from_dict(response.json())
        return {"logged": True, "info": response_data.to_dict()}
    else:
        return {"logged": False, "reason": "INVALID_API_TOKEN"}


def get_ai_messages(messages, stage, thread_id, headers: dict):
    url = f"{CLOUD_API_CLI_URL}/ai/messages"
    body = {
        "messages": messages,
        "runtime": stage,
        "threadId": thread_id,
        "version": "v1",
    }
    return requests.post(url, headers=headers, json=body, stream=True).iter_content(
        chunk_size=None
    )


def generate_project(prompt: str, abstra_json_version: str, headers: dict):
    url = f"{CLOUD_API_CLI_URL}/ai/generate"
    body = {"prompt": prompt, "version": abstra_json_version}
    r = requests.post(url, headers=headers, json=body)
    r.raise_for_status()
    return r.json()


def create_thread(headers: dict):
    url = f"{CLOUD_API_CLI_URL}/ai/thread"
    r = requests.post(url, headers=headers)
    r.raise_for_status()
    return r.json()


def cancel_all(headers: dict, thread_id: str):
    url = f"{CLOUD_API_CLI_URL}/ai/cancel-all"
    r = requests.post(url, headers=headers, json={"threadId": thread_id})
    r.raise_for_status()


def get_project_info(headers: dict, project_id: Optional[str] = None):
    if project_id:
        url = f"{CLOUD_API_CLI_URL}/project/{project_id}"
    else:
        url = f"{CLOUD_API_CLI_URL}/project"
    r = requests.get(url, headers=headers)
    r.raise_for_status()
    return r.json()


def get_project_id():
    if IS_PRODUCTION:
        return PROJECT_ID

    else:
        headers = resolve_headers()
        if headers is None:
            return None
        return get_project_info(headers)["id"]


class TunnelRequest(BaseModel):
    method: str
    path: str
    headers: dict
    body: Optional[str]
    query: dict
    sessionPath: str
    requestId: str


class TunnelResponse(BaseModel):
    status: int
    headers: dict
    text: str
    sessionPath: str
    requestId: str


class SessionPathMessage(BaseModel):
    sessionPath: str


def connect_tunnel():
    url = f"{CLOUD_API_ENDPOINT}/tunnel/connect".replace("https://", "wss://").replace(
        "http://", "ws://"
    )

    def loop():
        while True:
            headers = resolve_headers()
            if headers is not None:
                break
            sleep(1)
        ws = None
        while True:
            if ws is None:
                try:
                    ws = simple_websocket.Client(url, headers=headers)
                except Exception:
                    sleep(1)
                    continue
            try:
                message = ws.receive()
                assert isinstance(message, str)
                message_dict = json.loads(message)
                if "method" in message_dict:
                    request = TunnelRequest.model_validate_json(message)
                    kwargs: Any = dict(
                        headers=request.headers,
                        params=request.query,
                        **dict(data=request.body if request.body else {}),
                    )
                    if not request.path.startswith(
                        "/_hooks/"
                    ) and not request.path.startswith("/_tasks"):
                        response = TunnelResponse(
                            status=403,
                            headers={},
                            text=f"Forbidden path: {request.path}",
                            sessionPath=request.sessionPath,
                            requestId=request.requestId,
                        )
                        ws.send(response.model_dump_json())
                    else:
                        response = requests.request(
                            request.method,
                            f"http://{HOST}:{Settings.server_port}{request.path}",
                            **kwargs,
                        )
                        response_json = TunnelResponse(
                            status=response.status_code,
                            headers=dict(response.headers),
                            text=response.text,
                            sessionPath=request.sessionPath,
                            requestId=request.requestId,
                        ).model_dump_json()
                        ws.send(response_json)

                else:
                    global session
                    session = SessionPathMessage.model_validate_json(message)
                    public_url = (
                        f"{CLOUD_API_ENDPOINT}/tunnel/forward/{session.sessionPath}"
                    )
                    Settings.set_public_url(public_url)
                    print(
                        f"Hooks can also be fired from {Fore.GREEN} {public_url}/_hooks/:hook-path{Fore.RESET}"
                    )
            except simple_websocket.ConnectionClosed as e:
                print(f"Connection closed: {e}")
                ws = None
            except simple_websocket.ConnectionError as e:
                print(f"Connection error: {e}")
                ws = None
            except KeyboardInterrupt:
                ws.close()
                break
            except Exception as e:
                print(f"Error: {e}")
                ws.close()
                ws = None

    Thread(target=loop, daemon=True).start()



from abstra_internals.widgets.response_types import (
    AppointmentSlot,
    FileResponse,
    PhoneResponse,
)
from abstra_internals.widgets.types import AbstraOption

__all__ = [
    "AbstraOption",
    "FileResponse",
    "PhoneResponse",
    "AppointmentSlot",
]



import json
from dataclasses import dataclass
from typing import Any, Dict, List

from abstra_internals.cloud_api import (
    cancel_all,
    create_thread,
    generate_project,
    get_ai_messages,
)
from abstra_internals.controllers.linters import fix_all_linters
from abstra_internals.controllers.main import MainController
from abstra_internals.credentials import resolve_headers
from abstra_internals.repositories.project.json_migrations import get_latest_version
from abstra_internals.repositories.project.project import Project, ProjectRepository
from abstra_internals.settings import Settings


@dataclass
class PythonFile:
    filename: str
    content: str
    stage: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "PythonFile":
        return PythonFile(data["filename"], data["content"], data["stage"])


class AiController:
    def __init__(self, controller: MainController):
        self.controller = controller

    def send_ai_message(self, messages, stage, thread_id):
        headers = resolve_headers() or {}
        yield from get_ai_messages(messages, stage, thread_id, headers)

    def create_thread(self):
        headers = resolve_headers()
        if headers is None:
            return None
        return create_thread(headers)

    def cancel_all(self, thread_id: str):
        if headers := resolve_headers():
            cancel_all(headers, thread_id)

    def generate_project(self, prompt: str, tries: int = 0):
        headers = resolve_headers() or {}
        abstra_json_version = get_latest_version()
        try:
            res = generate_project(prompt, abstra_json_version, headers)
            is_abstra_json_valid = Project.validate(res["abstra_json"])
            if not is_abstra_json_valid:
                raise Exception("Generated abstra.json is not valid")

            generated_abstra_json = json.dumps(res["abstra_json"], indent=4)
            python_files = [PythonFile.from_dict(file) for file in res["python_files"]]

            self.init_stages(python_files)
            Settings.root_path.joinpath("abstra.json").write_text(
                generated_abstra_json, encoding="utf-8"
            )

            ProjectRepository.initialize_or_migrate(verbose=False)
            fix_all_linters()
        except Exception as e:
            if tries >= 3:
                raise e
            return self.generate_project(prompt, tries + 1)

    def init_stages(self, python_files: List[PythonFile]):
        for file in python_files:
            if file.stage == "form":
                script = self.controller.create_form(file.filename[:-3], file.filename)
            elif file.stage == "hook":
                script = self.controller.create_hook(file.filename[:-3], file.filename)
            elif file.stage == "script":
                script = self.controller.create_script(
                    file.filename[:-3], file.filename
                )
            elif file.stage == "job":
                script = self.controller.create_job(file.filename[:-3], file.filename)
            else:
                raise Exception(f"Invalid stage {file.stage}")
            script.file_path.write_text(file.content, encoding="utf-8")



import flask

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution_context import (
    HookContext,
    Response,
    extract_flask_request,
)
from abstra_internals.repositories.project.project import HookStage
from abstra_internals.usage import editor_usage
from abstra_internals.utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_hooks", __name__)

    @bp.route("/<path:id>", methods=["GET"])
    @editor_usage
    def _get_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)
        return hook.editor_dto

    @bp.route("/", methods=["GET"])
    @editor_usage
    def _get_hooks():
        return [f.editor_dto for f in controller.get_hooks()]

    @bp.route("/", methods=["POST"])
    @editor_usage
    def _create_hook():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        hook = controller.create_hook(title, file, workflow_position)
        return hook.editor_dto

    @bp.route("/<path:id>", methods=["PUT"])
    @editor_usage
    def _update_hook(id: str):
        changes = flask.request.json
        if not changes:
            flask.abort(400)

        hook = controller.update_stage(id, changes)
        if isinstance(hook, HookStage):
            return hook.editor_dto
        else:
            return None

    @bp.route("/<path:id>", methods=["DELETE"])
    @editor_usage
    def _delete_hook(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_hook(id, remove_file)
        return {"success": True}

    @bp.route("/<path:id>/run", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    @editor_usage
    def _run_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)

        context = HookContext(
            request=extract_flask_request(flask.request),
            response=Response(
                body="",
                headers={},
                status=200,
            ),
        )

        client = HookClient(context=context)

        ExecutionController(
            repositories=controller.repositories,
        ).run(
            stage=hook,
            client=client,
            context=context,
        )

        if context.response is None or client.context.response is None:
            flask.abort(500)

        return {
            "body": client.context.response.body,
            "status": context.response.status,
            "headers": context.response.headers,
        }

    return bp



from abstra_internals.widgets.widget_base import Input


class CodeInput(Input):
    type = 'code-input'
    empty_value = ''

    def __init__(self, key: str, label: str, **kwargs):
        initial_error = kwargs.get('errors', [])
        if not isinstance(initial_error, list):
            initial_error = [initial_error]
        super().__init__(key, initial_error)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.value = props.get('initial_value', self.empty_value)
        self.required = props.get('required', True)
        self.language = props.get('language', None)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'value': self.serialize_value(), 'language': self.language,
            'required': self.required, 'hint': self.hint, 'fullWidth': self
            .full_width, 'disabled': self.disabled, 'errors': self.errors}

    def serialize_value(self) ->str:
        return self.value if self.value is not None else ''



from typing import Dict

from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.entities.execution_context import HookContext, Request, Response


class HookClient(ExecutionClient):
    context: HookContext

    def __init__(self, context: HookContext) -> None:
        self.context = context

    def handle_failure(self, e: Exception) -> None:
        self.set_response(500, "An exception occurred during execution.", {})

    def handle_success(self) -> None:
        pass

    def set_response(self, status: int, body: str, headers: Dict[str, str]) -> None:
        self.context.response = Response(
            status=status,
            body=body,
            headers=headers,
        )

    def get_request(self) -> Request:
        return self.context.request

    def handle_start(self, execution_id: str):
        pass



from typing import List, Optional, TypedDict

from abstra_internals.entities.forms.form_state import (
    AutoNext,
    BackButton,
    Button,
    NextButton,
    State,
)
from abstra_internals.entities.forms.steps import (
    EndPageStep,
    GeneratorPageStep,
    PageStep,
    Step,
)
from abstra_internals.entities.forms.template import TemplateRenderer

BACK_ACTION = BackButton().label
NEXT_ACTION = NextButton().label
DEFAULT_ACTIONS = [NextButton().label, BackButton().label]


class StepsInfo(TypedDict):
    current: int
    total: int
    disabled: Optional[bool]


class RenderedForm(TypedDict):
    widgets: List[dict]
    end_page: bool
    steps_info: StepsInfo
    buttons: Optional[List[Button]]
    yielding: bool


class FormEntity:
    steps: List[Step]
    current_step_idx: int
    state: State
    hide_steps: bool

    def __init__(self, steps: List[Step], state: State, hide_steps: bool = False):
        self.steps = steps
        self.state = state
        self.hide_steps = hide_steps or False
        self.current_step_idx = 0

    def get_previous_page_idx(self) -> int:
        return next(
            self.current_step_idx - 1 - idx
            for idx, s in enumerate(self.steps[: self.current_step_idx][::-1])
            if not s.computation_step
        )

    def run(self) -> Optional[RenderedForm]:
        steps_info = StepsInfo(
            current=self.current_step_idx + 1,
            total=len(list(filter(lambda s: not s.computation_step, self.steps))),
            disabled=self.hide_steps,
        )

        if self.current_step_idx >= len(self.steps):
            return None

        step = self.steps[self.current_step_idx]

        if isinstance(step, PageStep):
            result = step.call(self.state)
            if result is None:
                self.current_step_idx += 1
                return self.run()
            if isinstance(result, AutoNext):
                self.current_step_idx += 1
                return self.run()
            if isinstance(result, tuple):
                template, buttons = result
            else:
                template, buttons = result, None
            renderer = TemplateRenderer(template)
            output = renderer.render(self.state)
            init_state = renderer.parse_state({})
            self.state.init_keys(init_state)
            end_page = isinstance(step, EndPageStep)
            self.has_errors = output["has_errors"]

            return RenderedForm(
                widgets=output["widgets"],
                end_page=end_page,
                steps_info=steps_info,
                buttons=buttons,
                yielding=False,
            )

        elif isinstance(step, GeneratorPageStep):
            generator = step.initialize_generator(self.state)
            buttons = None

            try:
                result = next(generator)
                if isinstance(result, tuple):
                    template, buttons = result
                else:
                    template, buttons = result, None
                renderer = TemplateRenderer(template)
                output = renderer.render(self.state)
                init_state = renderer.parse_state({})
                self.state.init_keys(init_state)
                self.has_errors = output["has_errors"]

                return RenderedForm(
                    widgets=output["widgets"],
                    end_page=False,
                    steps_info=steps_info,
                    buttons=buttons,
                    yielding=True,
                )

            except StopIteration:
                self.current_step_idx += 1

                if self.current_step_idx < len(self.steps):
                    return self.run()

                return None

        else:
            raise Exception(f"Internal error: invalid step type {step}")

    def handle_navigation(self, dto: dict) -> None:
        step = self.steps[self.current_step_idx]

        if not isinstance(step, PageStep) and not isinstance(step, GeneratorPageStep):
            raise Exception(f"Internal error: navigation on non-page step {step}")

        if dto["action"] not in DEFAULT_ACTIONS:
            self.state[dto["action"]] = True
            return

        if dto["action"] == BACK_ACTION:
            if self.current_step_idx == 0:
                raise Exception("Internal error: reached negative step index")
            self.current_step_idx = self.get_previous_page_idx()
            return

        result = step.call(self.state)
        if isinstance(result, tuple):
            template, _ = result
        else:
            template = result
        renderer = TemplateRenderer(template)

        self.state.merge(renderer.parse_state(dto["payload"]))
        self.state.touch_all()

        output = renderer.render(self.state)

        if output["has_errors"]:
            return

        self.current_step_idx += 1

    def handle_input(self, dto: dict) -> None:
        step = self.steps[self.current_step_idx]

        if not isinstance(step, PageStep) and not isinstance(step, GeneratorPageStep):
            raise Exception(f"Internal error: input on non-page step {step}")

        result = step.call(self.state)
        if isinstance(result, tuple):
            template, _ = result
        else:
            template = result

        renderer = TemplateRenderer(template)

        self.state.merge(renderer.parse_state(dto["payload"]))



import flask

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution_context import ScriptContext
from abstra_internals.repositories.project.project import ScriptStage
from abstra_internals.usage import editor_usage
from abstra_internals.utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_scripts", __name__)

    @bp.get("/<path:id>")
    @editor_usage
    def _get_script(id: str):
        script = controller.get_script(id)
        if not script:
            flask.abort(404)
        return script.editor_dto

    @bp.get("/")
    @editor_usage
    def _get_scripts():
        return [f.editor_dto for f in controller.get_scripts()]

    @bp.post("/")
    @editor_usage
    def _create_script():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        script = controller.create_script(title, file, workflow_position)
        return script.editor_dto

    @bp.put("/<path:id>")
    @editor_usage
    def _update_script(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        script = controller.update_stage(id, data)
        if isinstance(script, ScriptStage):
            return script.editor_dto
        else:
            return None

    @bp.delete("/<path:id>")
    @editor_usage
    def _delete_script(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_script(id, remove_file)
        return {"success": True}

    @bp.post("/<path:id>/run")
    @editor_usage
    def _run_script(id: str):
        script = controller.get_script(id)
        if not script:
            flask.abort(400)

        data = flask.request.json
        if not data:
            flask.abort(400)
        task_id = data.get("task_id")
        if not task_id:
            flask.abort(400)

        ExecutionController(
            repositories=controller.repositories,
        ).run(
            stage=script,
            context=ScriptContext(task_id=task_id),
        )

        return {"ok": True}

    return bp



from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.repositories.connectors import AccessTokenDTO

from .google_credentials import create_abstra_google_credentials


def get_access_token(name: str) -> AccessTokenDTO:
    context = SDKContextStore.get_by_thread()
    return context.repositories.connectors.get_access_token(name)


def get_gsheets_credentials(connection_name: str = "google-sheets"):
    return create_abstra_google_credentials(connection_name)


def get_gdrive_credentials(connection_name: str = "google-drive"):
    return create_abstra_google_credentials(connection_name)


def get_gmail_credentials(connection_name: str = "gmail"):
    return create_abstra_google_credentials(connection_name)


def get_bigquery_credentials(connection_name: str = "google-bigquery"):
    return create_abstra_google_credentials(connection_name)


def get_google_cal_credentials(connection_name: str = "google-calendar"):
    return create_abstra_google_credentials(connection_name)



# flake8: noqa
# Can't use __all__ yet because of the wildcard imports and code generation

from abstra_internals.interface.sdk.forms.generated.inputs import *
from abstra_internals.interface.sdk.forms.generated.outputs import *
from abstra_internals.interface.sdk.forms.list_item_schema import ListItemSchema
from abstra_internals.interface.sdk.forms.other import (
    execute_js,
    get_query_params,
    get_user,
    redirect,
    url_params,
)
from abstra_internals.interface.sdk.forms.deprecated.page import Page
from abstra_internals.interface.sdk.forms.deprecated.reactive_func import reactive
from abstra_internals.interface.sdk.forms.deprecated.step import run_steps
from abstra_internals.entities.forms.form_state import (
    Button,
    NextButton,
    BackButton,
    AutoNext,
)

from abstra_internals.widgets.library import *
from abstra_internals.interface.sdk.forms.form import Form, run
from abstra_internals.interface.sdk.forms.decorators import (
    end_page_step,
)



from typing import TYPE_CHECKING, Dict, List, Optional

from abstra_internals.entities.forms.form_entity import (
    FormEntity,
    Step,
)
from abstra_internals.entities.forms.form_state import State
from abstra_internals.interface.contract.forms_contract import (
    StepsInfo,
    ValidationResult,
)
from abstra_internals.interface.sdk import user_exceptions
from abstra_internals.jwt_auth import UserClaims
from abstra_internals.repositories.users import UsersRepository

if TYPE_CHECKING:
    from abstra_internals.controllers.execution_client_form import FormClient


class FormSDKController:
    def __init__(self, client: "FormClient", users_repository: UsersRepository) -> None:
        self.client = client
        self.users_repository = users_repository

    def get_user(self, force_refresh: bool) -> UserClaims:
        data = self.client.request_auth(force_refresh)
        claims = UserClaims.from_jwt(data["jwt"])

        if not claims:
            self.client.handle_invalid_jwt()
            raise user_exceptions.GetUserFailed()

        if user := self.users_repository.get_user(claims.email):
            claims.add_roles(user.roles)

        self.client.handle_valid_jwt()
        return claims

    def execute_js(self, code: str, context: dict):
        return self.client.request_execute_js(code, context)

    def redirect(self, url: str, query_params: Optional[Dict[str, str]] = None) -> None:
        self.client.request_redirect(url, query_params)

    def get_query_params(self) -> Dict[str, str]:
        return self.client.get_query_params()

    def next_user_message(self) -> Dict:
        return self.client.next_user_message()

    def run_form(self, steps: List[Step], state: State, hide_steps: bool) -> Dict:
        form = FormEntity(steps, state, hide_steps)
        rendered = None
        seq = 0

        while rendered := form.run():
            self.client.request_render(rendered, seq)

            if rendered["yielding"]:
                continue

            if rendered["end_page"]:
                break

            response = self.client.next_user_message()

            if "seq" in response:
                seq = response["seq"]

            if response["type"] == "form:input":
                form.handle_input(response)
            elif response["type"] == "form:navigation":
                form.handle_navigation(response)
            else:
                raise ValueError(f"Invalid response type: {response['type']}")

        return form.state.value

    def request_mount_page(
        self,
        widgets: list,
        actions: list,
        end_program: bool,
        reactive_polling_interval: int,
        steps_info: Optional[StepsInfo],
    ):
        self.client.request_mount_page(
            widgets,
            actions,
            end_program,
            reactive_polling_interval,
            steps_info,
        )

    # Deprecated by Forms v2
    def request_page_update(
        self, widgets: list, validation: ValidationResult, event_seq: int
    ):
        self.client.request_page_update(widgets, validation, event_seq)



import types
from typing import Any, Dict, List, Optional, Union, Tuple

from abstra_internals.entities.forms.form_state import State, Button
from abstra_internals.entities.forms.template import Template, TemplateFunc, TemplateRenderer
from abstra_internals.widgets.widget_base import Input


class LegacyListInput(Input):
    type = 'list-input'
    schemas: List[Any]
    item_schema: Any
    empty_value = []

    def __init__(self, key: str, item_schema: Any, **kwargs):
        initial_error = kwargs.get('errors', [])
        if not isinstance(initial_error, list):
            initial_error = [initial_error]
        super().__init__(key, initial_error)
        self.item_schema = item_schema
        self.schemas = []
        self.value = self.empty_value
        self.set_props(dict(item_schema=item_schema, **kwargs))

    def serialize_value(self) -> List:
        return [schema.serialize_value() for schema in self.schemas]

    def set_props(self, props):
        self.min = props.get('min', 0)
        self.max = props.get('max', None)
        self.hint = props.get('hint', None)
        self.add_button_text = props.get('add_button_text', '+')
        self.full_width = props.get('full_width', False)
        self.required = props.get('required', True)
        self.disabled = props.get('disabled', False)
        self.schemas = [self.item_schema.copy() for _ in range(self.min)]
        if props.get('initial_value', None):
            self.set_value(props.get('initial_value', None), set_errors=False)

    def _validate_list_min_max(self) -> List[str]:
        if self.min is not None and len(self.value) < self.min:
            return ['i18n_error_min_list']
        if self.max is not None and len(self.value) > self.max:
            return ['i18n_error_max_list']
        return []

    def _get_inputs(self):
        inputs = []
        for schema in self.schemas:
            inputs.extend(schema.get_input_widgets())
        return inputs

    def _validate_each(self) ->List[str]:
        inputs = self._get_inputs()
        for input in inputs:
            input.set_errors()
        if self._has_errors():
            return ['i18n_error_invalid_list_item']
        return []

    def _has_errors(self) ->bool:
        inputs = self._get_inputs()
        for input in inputs:
            has_errors = len(input.validate()) > 0
            if has_errors:
                return True
        return False

    @property
    def validators(self):
        return super().validators + [self._validate_list_min_max] + [self.
            _validate_each]

    def render(self, ctx: dict):
        schemas = [schema.render(ctx) for schema in self.schemas]
        return {'type': self.type, 'key': self.key, 'hint': self.hint,
            'errors': self.errors, 'min': self.min, 'max': self.max,
            'addButtonText': self.add_button_text, 'fullWidth': self.
            full_width, 'required': self.required, 'disabled': self.
            disabled, 'schemas': schemas, 'value': self.serialize_value()}

    def set_value(self, value: List, set_errors=False):
        if not self._has_errors():
            self.errors = []
        if not value:
            value = []
        self.value = value
        for idx, value_item in enumerate(value):
            if idx >= len(self.schemas):
                schema = self.item_schema.copy()
                self.schemas.append(schema)
            else:
                schema = self.schemas[idx]
            schema.set_values(value_item)
        self.schemas = self.schemas[:len(value)]

    def parse_value(self, value) -> List:
        parsed_values = []
        for index, value in enumerate(value or self.empty_value):
            if index >= len(self.schemas):
                schema = self.item_schema.copy()
                value = schema.parse_value(value or {})
            else:
                schema = self.schemas[index]
                value = schema.parse_value(value or {})
            parsed_values.append(value)
        return parsed_values


class _ListInput(Input):
    type = 'list-input'
    _raw_template: Union[Template, TemplateFunc]
    empty_value = []
    value: List[State]

    def __init__(self, key: str, raw_template: Union[Template, TemplateFunc], **kwargs):
        initial_error = kwargs.get('errors', [])
        if not isinstance(initial_error, list):
            initial_error = [initial_error]
        super().__init__(key, initial_error)
        self.value = self.empty_value
        self._raw_template = raw_template
        self.set_props(dict(**kwargs))

    def serialize_value(self):
        raise NotImplementedError("Unnecessary as its done by rendering widgets")


    def get_item_template(self, idx: int) -> Template:
        item_state = self.value[idx] if idx < len(self.value) else {}
        if callable(self._raw_template):
            result = self._raw_template(State(item_state))
            if isinstance(result, tuple):
                return result[0]
            return result
        return self._raw_template

    def set_props(self, props):
        self.min = props.get('min', 0)
        self.max = props.get('max', None)
        self.hint = props.get('hint', None)
        self.add_button_text = props.get('add_button_text', '+')
        self.full_width = props.get('full_width', False)
        self.required = props.get('required', True)
        self.disabled = props.get('disabled', False)

    def _validate_list_min_max(self) ->List[str]:
        if self.min is not None and len(self.value) < self.min:
            return ['i18n_error_min_list']
        if self.max is not None and len(self.value) > self.max:
            return ['i18n_error_max_list']
        return []

    def _validate_all_items(self) -> List[str]:
        for idx in range(len(self.value)):
            template = self.get_item_template(idx)
            renderer = TemplateRenderer(template)
            output = renderer.render(self.value[idx])
            if output["has_errors"]:
                return ['i18n_error_invalid_list_item']
        return []

    @property
    def validators(self):
        return super().validators + [self._validate_list_min_max, self._validate_all_items]
    


    def render(self, ctx: dict):
        rendered_items = []
        serialized_value = []
        for idx in range(len(self.value)):
            item_template = self.get_item_template(idx)
            renderer = TemplateRenderer(item_template)
            output = renderer.render(self.value[idx])
            rendered_items.append(output["widgets"])
            item_value = {widget['key']: widget['value'] for widget in output["widgets"] if 'key' in widget}
            serialized_value.append(item_value)
        
        return {
            'type': self.type,
            'key': self.key,
            'hint': self.hint,
            'errors': self.errors,
            'min': self.min,
            'max': self.max,
            'addButtonText': self.add_button_text,
            'fullWidth': self.full_width,
            'required': self.required,
            'disabled': self.disabled, 
            'schemas': rendered_items,
            'value': serialized_value,
        }
    
    def parse_value(self, value: List[Optional[Dict]]) -> List[State]:
        parsed_values: List[State] = []

        if not value:
            return self.empty_value
        
        for idx, item in enumerate(value):
            template = self.get_item_template(idx)
            renderer = TemplateRenderer(template)
            raw_value = item or {}
            parsed = renderer.parse_state(raw_value)
            parsed_values.append(parsed)
        
        return parsed_values

class ListInput:
    def __new__(cls, key: str, item_schema: Any, **kwargs):
        if isinstance(item_schema, list) or isinstance(item_schema, types.FunctionType):
            return _ListInput(key, raw_template=item_schema, **kwargs)
        return LegacyListInput(key, item_schema, **kwargs)


import io
import pathlib
import shutil
from typing import TYPE_CHECKING, Union

import requests

if TYPE_CHECKING:
    from PIL.Image import Image

from abstra_internals.constants import get_uploads_dir
from abstra_internals.utils.file import (
    get_random_filepath,
    get_tmp_upload_dir,
    upload_file,
)


def upload_widget_file(file: Union[str, io.IOBase, pathlib.Path, "Image"]) -> str:
    if not file:
        return ""

    if isinstance(file, (io.IOBase, pathlib.Path, str)):
        return upload_file(file)

    from PIL.Image import Image

    if isinstance(file, Image):
        _, file_path = get_random_filepath()
        file.save(str(file_path))
        return upload_file(open(file_path, "rb"))

    # FileResponse. TODO: check with isinstance without circular import
    if hasattr(file, "file"):
        return upload_file(file.file)

    raise ValueError(f"Cannot convert {type(file)}")


def download_to_path(url: str) -> pathlib.Path:
    # TODO: circular import - god help me
    from abstra_internals.controllers.sdk_context import (
        SDKContextStore,
    )

    execution = SDKContextStore.get_execution()
    save_dir = get_uploads_dir() / execution.id
    save_dir.mkdir(parents=True, exist_ok=True)
    save_path = save_dir / url.split("/")[-1]

    if url.startswith("http://") or url.startswith("https://"):
        with save_path.open("wb") as f, requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

        return save_path

    elif url.startswith("/_files/"):
        tmp_name = url[len("/_files/") :]
        path = get_tmp_upload_dir() / tmp_name
        shutil.copy(path, save_path)

        return save_path

    raise ValueError(f"Cannot download {url}")



from typing import List, Union

import abstra.ai as ai
import abstra.forms as af
from abstra.tasks import send_task
from abstra_internals.widgets.response_types import FileResponse

### 📌 Abstra Forms is the easiest way to create dynamic forms that allow users to interact with your Python script.

# 💡 Use read functions to receive input types, like multiple choice, text, boolean, and more (check the docs for the full list):
best_movie = af.read_multiple_choice(
    "What is the best movie ever made?",
    options=[
        "Die Hard",
        "Devil Wears Prada",
        "Weekend at Bernie's",
        "American Psycho",
        "Legally Blond",
        "Wall-E",
        "Threat Level: Midnight",
    ],
)

# 💡 Use Python to control logic and the user interactions:
if best_movie == "Threat Level: Midnight":
    af.display("*Michael Scarn theme song plays* Congratulations! You're a genius.")
elif best_movie == "Die Hard":
    af.display("Yep, that's the one. Passed the vibe check.")
else:
    af.display(
        "Nope. The correct answer was 'Die Hard'. Not off to a great start here."
    )

# 💡 Explore more input widgets like read_toggle to get a True/False response:
interview_opt_in = af.read_checkbox(
    "Do you agree to participate in a interview with Prison Mike?", required=False
)

# 💡 Or even read_camera function to prompt the user to submit a photo:
object_photo = af.read_camera(
    "Quick, take a photo of the object on your left. Do it now! 📸"
)


def process_photo(photo: Union[FileResponse, List[FileResponse]]) -> FileResponse:
    if isinstance(photo, list):
        if len(photo) > 0:
            return photo[0]
        else:
            exit("No photo was submitted.")
    else:
        return photo


### 📌 Abstra AI is a powerful tool that allows you to prompt an AI model to extract data from images, generate text and more

# 💡 Use the prompt function to send your prompt, attach files, add instructions and specify the response format:
generated_joke = ai.prompt(
    [
        "Write the punchline that Michael Scott from Dunder Mifflin would say if he saw this object on someone's desk in their office. Make sure the joke includes a quirky, slightly deprecating derogatory observation.",
        process_photo(object_photo),
    ],
    instructions="Only the punchline, and keep it office-friendly!",
)

rating = af.read_rating(f"Rate this joke's funniness: {generated_joke}", char="🤡")

### 📌 Abstra Workflows is a powerful tool that allows you to store and share data between different stages of your workflow

# 💡 Use the send_task function to send data to be used in the next stages:
send_task(
    "data_collector",
    {
        "best_movie": best_movie,
        "interview_opt_in": interview_opt_in,
        "rating": rating,
    },
)



import gc
import traceback
from pathlib import Path
from typing import Literal, Optional, Tuple

from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.controllers.execution_client_form import ClientAbandoned
from abstra_internals.controllers.sdk_context import SDKContext
from abstra_internals.entities.execution import Execution
from abstra_internals.logger import AbstraLogger
from abstra_internals.modules import import_as_new
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import StageWithFile
from abstra_internals.usage import execution_usage
from abstra_internals.utils.datetime import now_str

DEFAULT_STATUS = "failed"


def ExecutionTarget(
    *,
    stage: StageWithFile,
    execution: Execution,
    client: ExecutionClient,
    repositories: Repositories,
):
    with SDKContext(execution, client, repositories):
        status = DEFAULT_STATUS
        try:
            repositories.execution.create(execution)
            print(f"[ABSTRA] {now_str()} - Execution started")

            client.handle_start(execution.id)

            status, exception = _execute_code(stage.file_path)
            if exception:
                client.handle_failure(exception)
            else:
                client.handle_success()
        except ClientAbandoned:
            status = "abandoned"
        except Exception as e:
            status = "failed"
            AbstraLogger.capture_exception(e)
        finally:
            execution.teardown_tests()
            print(f"[ABSTRA] {now_str()} - Execution {status}")
            try:
                execution.set_status(status)
                repositories.execution.update(execution)
            except Exception as e_final:
                AbstraLogger.capture_exception(e_final)


def _execute_without_exit(filepath: Path):
    try:
        import_as_new(filepath.as_posix())
    except SystemExit as e:
        no_errors = e.code is None or e.code == 0
        if no_errors:
            return
        raise Exception(f"SystemExit: {e.code}") from e


Result = Tuple[Literal["finished", "abandoned", "failed"], Optional[Exception]]


def print_exception(exception: Exception, entrypoint: Path):
    tb = exception.__traceback__

    abstra_entrypoint_index = 0
    legacy_thread_index = 0
    index = 0

    while tb:
        # find abstra entrypoint
        if str(entrypoint).lower() == tb.tb_frame.f_code.co_filename.lower():
            abstra_entrypoint_index = index

        # find legacy threads, if used
        if tb.tb_frame.f_code.co_name == "use_legacy_threads":
            legacy_thread_index = index
        index += 1
        tb = tb.tb_next

    user_frame_index = max(abstra_entrypoint_index, legacy_thread_index) + 1

    # reset traceback
    tb = exception.__traceback__

    # remove unwanted frames
    for _ in range(user_frame_index):
        if tb:
            tb = tb.tb_next

    traceback.print_exception(type(exception), exception, tb)


@execution_usage
def _execute_code(filepath: Path) -> Result:
    try:
        _execute_without_exit(filepath)
        return "finished", None
    except ClientAbandoned as e:
        return "abandoned", e
    except Exception as e:
        print_exception(e, entrypoint=filepath)
        return "failed", e
    finally:
        gc.collect()



from abstra_internals.widgets.widget_base import Output


class LinkOutput(Output):
    type = 'link-output'

    def __init__(self, link_url: str, **kwargs):
        self.set_props(dict(link_url=link_url, **kwargs))

    def set_props(self, props):
        self.link_url = props.get('link_url', '')
        self.link_text = props.get('link_text', 'Click here')
        self.same_tab = props.get('same_tab', False)
        self.full_width = props.get('full_width', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'linkText': self.link_text, 'linkUrl':
            self.link_url, 'sameTab': self.same_tab, 'fullWidth': self.
            full_width}



import typing

from abstra_internals.widgets.metadata_parsed import WBrokerAPIParams, metadata_objects


def types_compatible(prop, typeName: typing.Union[typing.List[str], str]):
    if isinstance(typeName, list):
        return any(types_compatible(prop, tn) for tn in typeName)

    if prop is None:
        return True

    if typeName == "string":
        return isinstance(prop, str)
    if typeName == "number":
        return isinstance(prop, int) or isinstance(prop, float)
    if typeName == "boolean":
        return isinstance(prop, bool)
    if typeName == "array":
        return isinstance(prop, list)
    if typeName == "object":
        return isinstance(prop, dict)
    if typeName == "null":
        return prop is None
    if typeName == "any":
        return True

    raise Exception(f"Unknown type: {typeName}")


def valid_prop(prop, schema: WBrokerAPIParams):
    valid = True

    if schema.oneOf:
        valid = valid and prop in schema.oneOf

    return valid


def validate_widget_props(widget):
    metadata_widget = metadata_objects[widget["type"]]
    required_props = [prop for prop in metadata_widget.brokerAPI.params]

    for prop in required_props:
        prop_name = prop.argName
        if prop_name not in widget:
            raise Exception(
                f"Error in {widget['type']}: {prop_name} not in {widget.keys()}"
            )
        if not types_compatible(widget[prop_name], prop.typeName):
            raise Exception(
                f"{widget['type']}: {prop_name}: {type(widget[prop_name])} is not compatible with {prop}"
            )
        if not valid_prop(widget[prop_name], prop):
            raise Exception(f"{prop_name}: {widget[prop_name]} is not valid for {prop}")

    optional_props = [
        prop for prop in metadata_widget.brokerAPI.params if prop.isOptional
    ]
    for prop in optional_props:
        prop_name = prop.argName
        if not (
            widget.get(prop_name) is None
            or (
                types_compatible(widget[prop_name], prop.typeName)
                and valid_prop(widget[prop_name], prop)
            )
        ):
            raise Exception(
                f"{prop_name}: {type(widget[prop_name])} is not compatible with {prop}"
            )
    for prop in widget:
        if prop == "type":
            continue
        if not (prop in param.argName for param in metadata_widget.brokerAPI.params):
            raise Exception(f"{prop} not in {metadata_widget.brokerAPI.params}")



import pkg_resources
from packaging.version import Version


def get_local_package_version(package_name="abstra"):
    return parse_version(pkg_resources.get_distribution(package_name).version)


def parse_version(version: str):
    return Version(version)



import io
import json
import pathlib
from typing import Dict, List, Optional, Union

import pypdfium2 as pdfium
from PIL.Image import Image

import abstra_internals.utils.b64 as b64
from abstra_internals.repositories.ai import AiApiHttpClient
from abstra_internals.utils.string import to_snake_case
from abstra_internals.widgets.response_abc import AbstractFileResponse

Prompt = Union[str, io.IOBase, pathlib.Path, AbstractFileResponse]
Format = Dict[str, object]


class AiSDKController:
    def __init__(self, ai_client: AiApiHttpClient):
        self.ai_client = ai_client

    def _build_function_tool_call(self, format: Dict[str, object]) -> Dict[str, object]:
        variable_names = list(format.keys())
        description = "get_" + "_and_".join(variable_names)
        return {
            "name": "get_parameters",
            "description": to_snake_case(description),
            "parameters": {
                "type": "object",
                "properties": format,
                "required": variable_names,
            },
        }

    def _extract_pdf_images(self, file: Prompt) -> List[io.BytesIO]:
        images = []
        for page in pdfium.PdfDocument(file):
            bitmap = page.render(
                scale=4,  # 288 dpi
                rotation=0,
            )
            pil_image = bitmap.to_pil()
            image_io = io.BytesIO()
            pil_image.save(image_io, format="png")
            images.append(image_io)
        return images

    def _make_image_url_message(self, url: str):
        return {
            "role": "user",
            "content": [
                {
                    "type": "image_url",
                    "image_url": {
                        "url": url,
                    },
                },
            ],
        }

    def _make_text_message(self, text: str):
        return {
            "role": "user",
            "content": text,
        }

    def _try_extract_images(self, input: io.IOBase) -> Optional[List[io.BytesIO]]:
        try:
            images = self._extract_pdf_images(input)
            return images
        except Exception:
            return None

    def _make_messages(self, prompt: Prompt) -> List:
        if isinstance(prompt, pathlib.Path):
            with prompt.open("rb") as f:
                if images := self._try_extract_images(f):
                    return [
                        self._make_image_url_message(b64.encode_base_64(image))
                        for image in images
                    ]

                encoded_str = b64.encode_base_64(f)
                return [self._make_image_url_message(encoded_str)]

        if isinstance(prompt, AbstractFileResponse):
            file = prompt.file
            if images := self._try_extract_images(file):
                return [
                    self._make_image_url_message(b64.encode_base_64(image))
                    for image in images
                ]

            encoded_str = b64.encode_base_64(file)
            return [self._make_image_url_message(encoded_str)]

        if isinstance(prompt, io.IOBase):
            prompt.seek(0)
            if images := self._try_extract_images(prompt):
                return [
                    self._make_image_url_message(b64.encode_base_64(image))
                    for image in images
                ]
            encoded_str = b64.encode_base_64(prompt)
            return [self._make_image_url_message(encoded_str)]

        if isinstance(prompt, str) and (
            b64.is_base_64(prompt) or prompt.startswith("http")
        ):
            if prompt.endswith(".pdf"):
                raise ValueError("PDF URLs are not supported")

            return [self._make_image_url_message(prompt)]

        try:
            if isinstance(prompt, str) and pathlib.Path(prompt).exists():
                with open(prompt, "rb") as f:
                    if images := self._try_extract_images(f):
                        return [
                            self._make_image_url_message(b64.encode_base_64(image))
                            for image in images
                        ]
                    encoded_str = b64.encode_base_64(f)
                    return [self._make_image_url_message(encoded_str)]
        except OSError:  # Path contructor can raise OSError on long strings
            pass

        if isinstance(prompt, Image):
            image_io = io.BytesIO()
            prompt.save(image_io, format="PNG")
            image_io.seek(0)
            encoded_str = b64.encode_base_64(image_io)
            return [self._make_image_url_message(encoded_str)]

        if isinstance(prompt, str):
            return [self._make_text_message(prompt)]

        raise ValueError(f"Invalid prompt: {prompt}")

    def prompt(
        self,
        prompts: List[Prompt],
        instructions: List[str],
        format: Optional[Format],
        temperature: float,
    ):
        messages = []

        for instruction in instructions:
            messages.append(
                {
                    "role": "system",
                    "content": instruction,
                }
            )

        for prompt in prompts:
            messages.extend(self._make_messages(prompt))

        tools = []
        if format:
            function = self._build_function_tool_call(format)
            tools.append({"type": "function", "function": function})

        response = self.ai_client.prompt(messages, tools, temperature)

        if response.get("error"):
            raise Exception(response["error"])

        if format:
            parameters_dict = response["tool_calls"][0]["function"]["arguments"]
            try:
                return json.loads(parameters_dict)
            except json.JSONDecodeError:
                raise Exception(f"Error parsing JSON: {parameters_dict}")

        return response["content"]



from .agent import LocalRoleAgentRepository, ProductionRoleAgentRepository
from .client import LocalRoleClientRepository, ProductionRoleClientRepository

__all__ = [
    "LocalRoleAgentRepository",
    "ProductionRoleAgentRepository",
    "LocalRoleClientRepository",
    "ProductionRoleClientRepository",
]



import json
from abc import ABC, abstractmethod
from typing import Any, List

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import SIDECAR_HEADERS


class AiApiHttpClient(ABC):
    def __init__(self, base_url: str) -> None:
        self.url = f"{base_url}/ai"

    @abstractmethod
    def prompt(self, messages: List[Any], tools: List[Any], temperature: float):
        raise NotImplementedError()


class ProductionAiApiHttpClient(AiApiHttpClient):
    def prompt(self, messages: List[Any], tools: List[Any], temperature: float):
        body = {
            "messages": messages,
            "tools": tools,
            "temperature": temperature,
        }
        response = requests.post(
            f"{self.url}/prompt", headers=SIDECAR_HEADERS, json=body
        )
        try:
            response = response.json()
            return response
        except json.JSONDecodeError:
            raise Exception(f"Error parsing JSON: {response.text}")


class LocalAiApiHttpClient(AiApiHttpClient):
    def prompt(self, messages: List[Any], tools: List[Any], temperature: float):
        body = {
            "messages": messages,
            "tools": tools,
            "temperature": temperature,
        }
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to use AI")
        response = requests.post(f"{self.url}/prompt", headers=headers, json=body)
        try:
            response = response.json()
            return response
        except json.JSONDecodeError:
            raise Exception(f"Error parsing JSON: {response.text}")



import re
from typing import List
from abstra_internals.widgets.widget_base import Input


class EmailInput(Input):
    type = 'email-input'
    empty_value = ''
    value: str

    def __init__(self, key: str, label: str, **kwargs):
        initial_error = kwargs.get('errors', [])
        if not isinstance(initial_error, list):
            initial_error = [initial_error]
        super().__init__(key, initial_error)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.value = props.get('initial_value', self.empty_value)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.placeholder = props.get('placeholder', '')
        self.full_width = props.get('full_width', False)
        self.invalid_email_message = props.get('invalid_email_message',
            'i18n_error_invalid_email')
        self.disabled = props.get('disabled', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'value': self.serialize_value(), 'placeholder': self.
            placeholder, 'required': self.required, 'hint': self.hint,
            'fullWidth': self.full_width, 'invalidEmailMessage': self.
            invalid_email_message, 'disabled': self.disabled, 'errors':
            self.errors}

    def serialize_value(self) ->str:
        return self.value if self.value else ''

    def validate(self) ->List[str]:
        errors = super().validate()
        if len(self.value) == 0:
            return errors
        is_valid_email = bool(re.search(
            '^(([^<>()[\\]\\\\.,;:\\s@"]+(\\.[^<>()[\\]\\\\.,;:\\s@"]+)*)|(".+"))@((\\[[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\])|(([a-zA-Z\\-0-9]+\\.)+[a-zA-Z]{2,}))$'
            , self.value.strip().lower()))
        if not is_valid_email:
            errors.append(self.invalid_email_message)
        return errors



from abstra.messages import send_email
from abstra.tasks import get_trigger_task, send_task

### Abstra Workflows is the easiest way to store and manage data across different parts of your workflow

# You can use the get_trigger_task function to get the task that triggered the script
task = get_trigger_task()
best_movie = task["best_movie"]
interview_opt_in = task["interview_opt_in"]
rating = task["rating"]

reasons = []

if best_movie != "Die Hard" and best_movie != "Threat Level: Midnight":
    reasons.append("Applicant does not have good cinematic taste.")

if not interview_opt_in:
    reasons.append(
        "Applicant refused to engage in a mandatory but voluntary interview with 'Prison Mike'."
    )

if rating != 5:
    reasons.append(
        "Applicant didn't laugh at Michael's joke, making Michael question his sense of humor."
    )

print(reasons)
### Abstra Messages is a built-in notification module that makes automating your processes much easier.

# Use send_email to send emails. All emails in development mode are sent to the editor's email (that's you!).
email = "name@email.com"  # this is a placeholder

if not reasons:
    send_email(
        email,
        title="Welcome to Dunder Mifflin!",
        message="Welcome to the team! You can now join as Secretary of the Assistant to the Regional Manager. \n\n See you Monday, \n\n Michael Scott - World's Best Boss.",
    )
else:
    reasons_text = "\n".join(reasons)
    print(reasons_text)
    send_email(
        email,
        title="Nope, Sorry!",
        message=f"Your application to work at Dunder Mifflin has been rejected. Here's why:\n {reasons_text} \n\n Sincerely Not Sorry, \n\n Michael Scott - World's Best Boss.",
    )

# Store the application response for future use:
if not reasons:
    send_task("approved", {})
else:
    send_task("rejected", {"reasons": reasons})

task.complete()



import json
from datetime import datetime
from pathlib import Path

import requests
from packaging.version import Version

from abstra_internals.logger import AbstraLogger

from .dot_abstra import CACHED_VERSIONS_DIR

EXPIRE_PERIOD = 60 * 60 * 4  # 4 hours
TIMEOUT = 5


def get_cached_latest_version(root_path: Path, package_name="abstra"):
    cached_file = root_path / CACHED_VERSIONS_DIR / f"{package_name}.json"

    if cached_file.exists():
        try:
            with open(cached_file, "r") as f:
                cached_version = json.loads(f.readline())

            created_at = cached_version["created_at"]
            version = cached_version["version"]
        except Exception as e:
            AbstraLogger.capture_exception(e)
            created_at = None
            version = None

        if (
            created_at is not None
            and isinstance(created_at, float)
            and version is not None
            and datetime.utcnow().timestamp() - created_at < EXPIRE_PERIOD
        ):
            return Version(version)

    try:
        response = requests.get(
            f"https://pypi.org/pypi/{package_name}/json", timeout=TIMEOUT
        )
        response.raise_for_status()
        latest_version = response.json()["info"]["version"]

        update_cached_latest_version(root_path, latest_version, package_name)

        return Version(latest_version)
    except Exception:
        return None


def update_cached_latest_version(
    root_path: Path, version: Version, package_name="abstra"
):
    cached_file = root_path / CACHED_VERSIONS_DIR / f"{package_name}.json"

    if not cached_file.parent.exists():
        cached_file.parent.mkdir(parents=True)

    with open(cached_file, "w") as f:
        json.dump(
            {
                "version": str(version),
                "created_at": datetime.utcnow().timestamp(),
            },
            f,
        )



from typing import Dict, List, Union

from abstra_internals.controllers.common.task_executors import TaskExecutor
from abstra_internals.entities.execution import Execution
from abstra_internals.entities.execution_context import ScriptContext
from abstra_internals.interface.sdk import user_exceptions
from abstra_internals.interface.sdk.user_exceptions import (
    TaskNotWaiting,
    TaskWriteAttempt,
)
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.repositories.tasks import TaskDTO, TaskLockFailed, TaskPayload


class Task:
    def __init__(self, task_controller: "TasksSDKController", dto: TaskDTO) -> None:
        self._dto = dto
        self._controller = task_controller

    def __str__(self) -> str:
        return str(self._dto)

    def __repr__(self) -> str:
        return self._dto.__repr__()

    def __getitem__(self, key: str) -> object:
        return self._dto.payload[key]

    def __setitem__(self, key: str, value: str) -> None:
        raise TaskWriteAttempt()

    @property
    def id(self) -> str:
        return self._dto.id

    @property
    def type(self) -> str:
        return self._dto.type

    @property
    def payload(self) -> TaskPayload:
        return self._dto.payload

    def get_payload(self) -> TaskPayload:
        return self._dto.payload

    def get_dto(self) -> TaskDTO:
        return self._dto

    def lock(self) -> None:
        try:
            self._controller.lock_task(self._dto.id)
        except TaskLockFailed:
            raise TaskNotWaiting()

    def complete(self) -> None:
        try:
            self._controller.complete_task(self._dto.id)
        except TaskLockFailed:
            raise TaskNotWaiting()

    def __enter__(self):
        self.lock()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.complete()
        else:
            self._controller.set_task_to_pending(self._dto.id)
            return False


class TasksSDKController:
    def __init__(self, repositories: Repositories, execution: Execution) -> None:
        self.execution = execution
        self.repositories = repositories
        self.project = ProjectRepository().load()
        self.executor = TaskExecutor(repositories)

    def send_task(self, type: str, payload: TaskPayload) -> None:
        stage = self.project.get_stage_raises(self.execution.stage_id)
        self.executor.send_task(
            type=type, current_stage=stage, payload=payload, execution=self.execution
        )

    def get_stage_pending_tasks(
        self, limit: Union[int, None], offset: int, where: Dict
    ) -> List[Task]:
        if self.execution.context.mock_execution.test_pending_tasks:
            return [
                Task(self, task)
                for task in self.execution.context.mock_execution.test_pending_tasks
            ]

        stage = self.project.get_stage(self.execution.stage_id)
        if not stage:
            raise Exception(f"Stage {self.execution.stage_id} not found")

        dtos = self.repositories.tasks.get_pending_tasks(
            self.execution.stage_id, limit, offset, where
        )
        return [Task(self, dto) for dto in dtos]

    def get_stage_sent_tasks(
        self, limit: Union[int, None], offset: int, where: Dict
    ) -> List[Task]:
        stage = self.project.get_stage(self.execution.stage_id)
        if not stage:
            raise Exception(f"Stage {self.execution.stage_id} not found")

        dtos = self.repositories.tasks.get_sent_tasks(
            self.execution.stage_id, limit, offset, where
        )
        return [Task(self, dto) for dto in dtos]

    def lock_task(self, task_id: str) -> None:
        self.repositories.tasks.lock_task(
            task_id, self.execution.id, self.execution.stage_id
        )

    def complete_task(self, task_id: str) -> None:
        self.repositories.tasks.complete_task(
            task_id, self.execution.id, self.execution.stage_id
        )

    def set_task_to_pending(self, task_id: str) -> None:
        self.repositories.tasks.set_task_to_pending(task_id)

    def get_all_tasks(self) -> List[Task]:
        dtos = self.repositories.tasks.get_all_tasks()
        return [Task(self, dto) for dto in dtos]

    def get_trigger_task(self) -> Task:
        if not isinstance(self.execution.context, ScriptContext):
            raise user_exceptions.IncompatibleScriptSDK()

        if self.execution.context.mock_execution.test_trigger_task:
            return Task(self, self.execution.context.mock_execution.test_trigger_task)

        dto = self.repositories.tasks.get_by_id(self.execution.context.task_id)
        return Task(self, dto)



import io
from typing import List, Union

from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.email_templates import message_template


def send_email(
    to: Union[str, List[str]],
    message: str,
    title: str = "",
    attachments: List[Union[str, io.IOBase]] = [],
):
    if isinstance(to, str):
        to = [to]
    SDKContextStore.get_by_thread().repositories.email.send(
        message_template.generate_email(to, message, title, attachments)
    )


__all__ = [
    "send_email",
]



import datetime
from typing import Generic, Literal, Optional, TypeVar
from uuid import uuid4

from pydantic import field_serializer

from abstra_internals.entities.execution_context import (
    ClientContext,
    FormExecutionMock,
    HookExecutionMock,
    ScriptExecutionMock,
)
from abstra_internals.utils.datetime import to_utc_iso_string
from abstra_internals.utils.serializable import Serializable

T = TypeVar("T", bound=ClientContext)
ExecutionStatus = Literal["running", "failed", "finished", "abandoned"]


class Execution(Serializable, Generic[T]):
    id: str
    stage_id: str
    status: ExecutionStatus
    created_at: datetime.datetime
    updated_at: Optional[datetime.datetime] = None
    context: T

    @classmethod
    def create(
        cls,
        *,
        stage_id: str,
        context: T,
    ) -> "Execution[T]":
        return cls(
            stage_id=stage_id,
            context=context,
            status="running",
            id=uuid4().__str__(),
            created_at=datetime.datetime.now(),
            updated_at=None,
        )

    def teardown_tests(self):
        self.context.mock_execution.test_pending_tasks = []
        if isinstance(self.context.mock_execution, HookExecutionMock):
            self.context.mock_execution.test_request = None
        elif isinstance(self.context.mock_execution, FormExecutionMock):
            self.context.mock_execution.test_answers = []
        elif isinstance(self.context.mock_execution, ScriptExecutionMock):
            self.context.mock_execution.test_trigger_task = None

    def set_status(self, status: ExecutionStatus) -> None:
        if status == "running":
            raise ValueError("Cannot set status to running")

        self.status = status
        self.updated_at = datetime.datetime.now()

    @property
    def short_id(self) -> str:
        return self.id[:8]

    @field_serializer("created_at")
    def serialize_created_at(self, value):
        return to_utc_iso_string(value)

    @field_serializer("updated_at")
    def serialize_updated_at(self, value):
        return to_utc_iso_string(value) if value is not None else None


class PreExecution(Serializable):
    stage_id: str
    context: Optional[ClientContext] = None



from abc import ABC, abstractmethod
from typing import Dict, List, Literal, Optional

from abstra_internals.entities.agents import AgentModel, ConnectionModel


class RoleCommonRepository(ABC):
    _cached_agents: Optional[Dict[str, AgentModel]] = None
    _cached_connections: Optional[List[ConnectionModel]] = None
    base_url: str

    @abstractmethod
    def get_connections(self) -> List[ConnectionModel]:
        raise NotImplementedError

    @property
    @abstractmethod
    def role(self) -> Literal["client", "agent"]:
        raise NotImplementedError

    def __init__(self, base_url: str) -> None:
        self.base_url = base_url + "/services/roles/" + self.role

    @abstractmethod
    def get_headers(self):
        raise NotImplementedError

    @abstractmethod
    def get_current_project_id(self):
        raise NotImplementedError



from typing import List, Literal

from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.entities.agents import ConnectionModel


def get_connections(role: Literal["client", "agent"]) -> List[ConnectionModel]:
    if role == "client":
        return (
            SDKContextStore.get_by_thread().repositories.role_clients.get_connections()
        )
    else:
        return (
            SDKContextStore.get_by_thread().repositories.role_agents.get_connections()
        )


__all__ = ["get_connections"]



import pkg_resources
from pkg_resources import get_distribution as gd

from abstra_internals.linter.rules.missing_abstra_in_requirements import (
    AbstraVersionInRequirementsIsBehindInstalled,
    MissingAbstraInRequirements,
)
from tests.fixtures import BaseTest


def mock_get_distribution(name):
    if name == "abstra":
        return pkg_resources.Distribution(project_name="abstra", version="1.0.0")
    else:
        return gd(name)


class MissingAbstraInRequirementsTest(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.old_get_distribution = pkg_resources.get_distribution
        pkg_resources.get_distribution = mock_get_distribution

    def tearDown(self) -> None:
        pkg_resources.get_distribution = self.old_get_distribution
        super().tearDown()

    def test_missing_abstra_in_requirements_valid_with_requirements(self):
        rule = MissingAbstraInRequirements()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("abstra==1.0.0")
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_abstra_in_requirements_invalid_with_different_version(self):
        rule = MissingAbstraInRequirements()
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("abstra==0.0.1")
        issues = rule.find_issues()
        self.assertEqual(len(issues), 1)
        self.assertIsInstance(issues[0], AbstraVersionInRequirementsIsBehindInstalled)

    def test_missing_abstra_in_requirements_invalid_without_requirements(self):
        rule = MissingAbstraInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_abstra_in_requirements_fix(self):
        rule = MissingAbstraInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)
        rule.find_issues()[0].fixes[0].fix()
        self.assertEqual(len(rule.find_issues()), 0)
        requirements_txt = self.root / "requirements.txt"
        self.assertTrue(requirements_txt.exists())
        with requirements_txt.open("r") as file:
            content = file.read()
            self.assertTrue("abstra==1.0.0" in content)



import hashlib

import jwt

from abstra_internals.environment import EMAIL_JWT_AUDIENCE, PUBLIC_KEY
from abstra_internals.logger import AbstraLogger


def decode_email_jwt_header(jwt_header: str):
    try:
        if PUBLIC_KEY:
            return jwt.decode(
                jwt_header,
                key=PUBLIC_KEY,
                algorithms=["RS256"],
                audience=EMAIL_JWT_AUDIENCE,
            )
        return jwt.decode(
            jwt_header, options={"verify_signature": False}, audience=EMAIL_JWT_AUDIENCE
        )

    except jwt.ExpiredSignatureError:
        return None
    except Exception as e:
        AbstraLogger.capture_exception(e)
        return None


def verify_email(*, jwt_header: str, content: str) -> bool:
    calculated_hash = hashlib.sha256(content.encode()).hexdigest()

    data = decode_email_jwt_header(jwt_header)
    if not data:
        return False

    return calculated_hash == data.get("hash")


__all__ = ["verify_email"]



import abc

import requests

from abstra_internals.environment import SIDECAR_HEADERS


class KVRepository(abc.ABC):
    def get(self, key: str):
        raise NotImplementedError()

    def delete(self, key: str):
        raise NotImplementedError()

    def set(self, key: str, value: str):
        raise NotImplementedError()


class ProductionKVRepository(KVRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/kv"

    def get(self, key: str):
        r = requests.get(f"{self.base_url}/{key}", headers=SIDECAR_HEADERS)
        if not r.ok:
            return None

        return r.json()["value"]

    def delete(self, key: str):
        requests.delete(
            f"{self.base_url}/{key}", headers=SIDECAR_HEADERS
        ).raise_for_status()

    def set(self, key: str, value: str):
        requests.put(
            f"{self.base_url}/{key}", json=dict(value=value), headers=SIDECAR_HEADERS
        ).raise_for_status()


class LocalKVRepository(KVRepository):
    kvs = {}

    def get(self, key: str):
        return self.kvs.get(key)

    def delete(self, key: str):
        del self.kvs[key]

    def set(self, key: str, value: str):
        self.kvs[key] = value




import io
from typing import Union, TYPE_CHECKING
from abstra_internals.interface.sdk.forms.deprecated.page import Page

if TYPE_CHECKING:
    from pandas import DataFrame
    from plotly.graph_objects import Figure


def display_file(file: Union[str, io.IOBase], **kwargs):
  
    """Show a button for the user to download a file.

        Position Args: 
		file (Union[str, io.IOBase]): The file to download
		
        Keyword Args: 
		download_text (str): The text to display on the button that will download the file. Defaults to "Download here".
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_file(file, **kwargs).run(button_text, end_program=end_program)

def display_html(html: str, **kwargs):
  
    """Simple HTML embed widget.

        Position Args: 
		html (str): The html snippet to display to the user
		
        Keyword Args: 
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_html(html, **kwargs).run(button_text, end_program=end_program)

def display_iframe(url_or_html: str, **kwargs):
  
    """Embed content using an iframe.

        Position Args: 
		url_or_html (str): The link to the document or the own document to display to the user
		
        Keyword Args: 
		width (int): The width of the iframe. Defaults to "800".
		height (int): The height of the iframe. Defaults to "600".
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_iframe(url_or_html, **kwargs).run(button_text, end_program=end_program)

def display_image(image: Union[str, io.IOBase], **kwargs):
  
    """Show an image to the user

        Position Args: 
		image (Union[str, io.IOBase]): The image to display to the user
		
        Keyword Args: 
		subtitle (str): The subtitle of the image. Defaults to "".
		label (str): The label to display to the user
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_image(image, **kwargs).run(button_text, end_program=end_program)

def display_latex(text: str, **kwargs):
  
    """Display a latex formula to the user.

        Position Args: 
		text (str): The latex formula to display to the user
		
        Keyword Args: 
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_latex(text, **kwargs).run(button_text, end_program=end_program)

def display_link(link_url: str, **kwargs):
  
    """Display a link to the user.

        Position Args: 
		link_url (str): The url of the link to display to the user
		
        Keyword Args: 
		link_text (str): The text to display on the link. Defaults to "Click here".
		same_tab (bool): Whether to open the link in the same tab or not. Defaults to False.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_link(link_url, **kwargs).run(button_text, end_program=end_program)

def display_markdown(text: str, **kwargs):
  
    """Show markdown formatted text to the user

        Position Args: 
		text (str): The formatted text to display to the user
		
        Keyword Args: 
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_markdown(text, **kwargs).run(button_text, end_program=end_program)

def display_pandas(df: "DataFrame", **kwargs):
  
    """Display a pandas dataframe to the user

        Position Args: 
		df ("DataFrame"): The dataframe to display to the user
		
        Keyword Args: 
		display_index (bool): Whether to show a index column. Defaults to False.
		label (str): The label to display to the user
		page_size (number): The number of rows to display per page. Defaults to 10.
		full_width (bool): Whether the input should use full screen width. Defaults to True.
		actions (list): Actions that can be triggered by table rows
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_pandas(df, **kwargs).run(button_text, end_program=end_program)

def display_plotly(fig: "Figure", **kwargs):
  
    """Displays a Plotly figure.

        Position Args: 
		fig ("Figure"): The figure to display to the user
		
        Keyword Args: 
		label (str): The label to display to the user
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_plotly(fig, **kwargs).run(button_text, end_program=end_program)

def display_progress(current: float, total: float, **kwargs):
  
    """Displays a progress bar to the user.

        Position Args: 
		current (float): The progress being made. Defaults to 50.
		total (float): Total progress. Defaults to 100.
		
        Keyword Args: 
		text (str): The text displayed with this progress step. Defaults to "".
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display_progress(current, total, **kwargs).run(button_text, end_program=end_program)

def display(text: str, **kwargs):
  
    """"Text display" adds textual information into the app.

        Position Args: 
		text (str): The text to display to the user
		
        Keyword Args: 
		size (str): The size of the text: 'small', 'medium', 'large'
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    end_program = kwargs.get("end_program", False)
    return Page().display(text, **kwargs).run(button_text, end_program=end_program)


__all__ = [
    "display_file",
    "display_html",
    "display_iframe",
    "display_image",
    "display_latex",
    "display_link",
    "display_markdown",
    "display_pandas",
    "display_plotly",
    "display_progress",
    "display"
]



import flask

from abstra_internals.controllers.main import ExecutionFilter, MainController
from abstra_internals.usage import editor_usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("executions", __name__)

    @bp.get("/")
    @editor_usage
    def _get_executions():
        filter = ExecutionFilter.from_dict(flask.request.args)
        executions = controller.get_executions(filter)
        return executions.to_dict()

    @bp.get("/stages")
    @editor_usage
    def _get_stages():
        stages = controller.get_stages()
        return [stage.as_dict for stage in stages]

    return bp



import unittest
from typing import Dict

from abstra_internals.entities.forms.form_entity import FormEntity
from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.steps import PageStep, Step
from abstra_internals.entities.forms.template import Template
from abstra_internals.widgets.library import ListInput, TextInput, TextOutput

rendered_text_input = {
    "type": "text-input",
    "key": "name",
    "label": "Name",
    "value": "",
    "placeholder": "",
    "required": True,
    "hint": None,
    "fullWidth": False,
    "mask": None,
    "maxLength": None,
    "minLength": None,
    "disabled": False,
    "errors": [],
}

rendered_text_output = {
    "type": "text-output",
    "text": "test",
    "size": "medium",
    "fullWidth": False,
}

rendered_list_input = {
    "type": "list-input",
    "key": "list",
    "hint": None,
    "errors": [],
    "min": 0,
    "max": None,
    "addButtonText": "+",
    "fullWidth": False,
    "required": True,
    "disabled": False,
    "schemas": [],
    "value": [],
}


def compare_renders(rendered, expected):
    assert rendered["end_page"] == expected["end_page"]
    assert rendered["steps_info"] == expected["steps_info"]
    assert len(rendered["widgets"]) == len(expected["widgets"])
    for i, widget in enumerate(rendered["widgets"]):
        for key, value in widget.items():
            if value != expected["widgets"][i][key]:
                assert False, f"Widget {i} key {key} does not match.\nEXPECTED:\n{expected['widgets'][i][key]}\nGOT\n{value}"


def make_steps() -> Dict[str, Step]:
    def input_page(state: State) -> Template:
        return [TextInput(key="name", label="Name")]

    def output_page(state: State) -> Template:
        return [TextOutput(str(state["name"]))]

    def reactive_page(state: State) -> Template:
        p: Template = [TextInput(key="name", label="Name")]
        if state.get("name") == "test":
            p.append(TextOutput(str(state["name"])))
        return p

    def list_input_schema(item_state: State) -> Template:
        t: Template = [TextInput(key="name", label="Name")]
        if item_state.get("name") == "test":
            t.append(TextOutput(str(item_state["name"])))
            t.append(TextInput(key="email", label="Email"))
        return t

    def list_page(state: State) -> Template:
        return [ListInput(item_schema=list_input_schema, key="list")]

    def list_input_page(state: State):
        return [
            TextInput(key="name", label="Name"),
            ListInput(item_schema=list_input_schema, key="list"),
        ]

    def compute(state):
        pass

    return {
        "input_page": PageStep(input_page),
        "output_page": PageStep(output_page),
        "reactive_page": PageStep(reactive_page),
        "list_page": PageStep(list_page),
        "list_input_page": PageStep(list_input_page),
    }


class FormEntityTest(unittest.TestCase):
    def test_input_page(self):
        all_steps = make_steps()
        steps = [all_steps["input_page"]]
        form = FormEntity(steps, State())
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "test"},
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "test",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

    def test_reactive_page(self):
        all_steps = make_steps()
        steps = [all_steps["reactive_page"]]
        form = FormEntity(steps, State())
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "test"},
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "test",
                    },
                    {
                        **rendered_text_output,
                    },
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "tes"},
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "tes",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

    def test_validation(self):
        all_steps = make_steps()
        steps = [all_steps["input_page"]]
        form = FormEntity(steps, State())
        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": "my name"},
            }
        )

        rendered = form.run()

        form.handle_input(
            {
                "type": "form:input",
                "payload": {"name": ""},
            }
        )
        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "",
                        "errors": ["i18n_error_required_field"],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

    def test_list_page(self):
        all_steps = make_steps()
        steps = [all_steps["list_page"]]
        form = FormEntity(steps, State())

        # Initial render
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_list_input,
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

        # Add item
        form.handle_input(
            {
                "type": "form:input",
                "payload": {"list": [None]},
            }
        )
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_list_input,
                        "schemas": [[rendered_text_input]],
                        "value": [{"name": ""}],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

        # Fill item
        form.handle_input(
            {
                "type": "form:input",
                "payload": {"list": [{"name": "test"}]},
            }
        )
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_list_input,
                        "schemas": [
                            [
                                {
                                    **rendered_text_input,
                                    "value": "test",
                                },
                                {
                                    **rendered_text_output,
                                    "text": "test",
                                },
                                {
                                    **rendered_text_input,
                                    "key": "email",
                                    "label": "Email",
                                    "value": "",
                                },
                            ]
                        ],
                        "value": [{"name": "test", "email": ""}],
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 1, "disabled": False},
            },
        )

    def test_navigation(self):
        all_steps = make_steps()
        steps = [all_steps["input_page"], all_steps["output_page"]]
        form = FormEntity(steps, State())
        rendered = form.run()
        compare_renders(
            rendered,
            {
                "widgets": [rendered_text_input],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_next_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_output,
                        "text": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 2, "total": 2, "disabled": False},
            },
        )

        form.handle_navigation(
            {
                "type": "form:navigation",
                "payload": {"name": "test-name"},
                "action": "i18n_back_action",
            }
        )

        rendered = form.run()

        compare_renders(
            rendered,
            {
                "widgets": [
                    {
                        **rendered_text_input,
                        "value": "test-name",
                    }
                ],
                "end_page": False,
                "steps_info": {"current": 1, "total": 2, "disabled": False},
            },
        )



from typing import List

from abstra_internals.constants import get_project_url
from abstra_internals.email_templates.html_wrapper import style_wrap
from abstra_internals.email_templates.i18n import get_translation
from abstra_internals.repositories.email import EmailParams
from abstra_internals.repositories.project.project import FormStage, ProjectRepository

template = """
<p style="font-style: normal; font-weight: 700; font-size: 30px; line-height: 36px; color:#181818; margin-top:30px;">
    {a_form_is_waiting}
</p>
<p style="margin-bottom: 40px; font-size: 16px;">New task</p>
<a href="{stage_link}" style="text-decoration: none; padding: 8px 20px; border-radius: 6px; border: 1px solid transparent; background-color: #d14056; color: #FFF; box-shadow: 0 2px 0 rgba(255, 5, 5, 0.06); width: fit-content; font-size: 16px; font-family: system-ui;">
    {waiting_cta}
</a>
"""


def generate_email(recipient_emails: List[str], form: FormStage) -> EmailParams:
    project = ProjectRepository.load()
    translations = get_translation(project.workspace.language or "en")
    stage_link = f"{get_project_url()}/{form.path}"

    content = template.format(
        a_form_is_waiting=translations.form_is_waiting(form.title),
        stage_link=stage_link,
        waiting_cta=translations.waiting_cta(),
    )

    html = style_wrap(content, project.workspace)

    return EmailParams(
        kind="task-waiting",
        to=recipient_emails,
        subject=translations.form_is_waiting(form.title),
        body=html,
        is_html=True,
        attachments=[],
    )



import sys
from typing import Callable, List, Literal, Union

import flask_sock

from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.sdk_context import (
    ExecutionNotFound,
    SDKContextStore,
)
from abstra_internals.env_masker import GLOBAL_MASKER
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.logger import AbstraLogger
from abstra_internals.utils import serialize


class StdioController:
    listeners: List[flask_sock.Server] = []

    @classmethod
    def register(cls, listener: flask_sock.Server):
        cls.listeners.append(listener)

    @classmethod
    def unregister(cls, listener: flask_sock.Server):
        cls.listeners.remove(listener)

    @classmethod
    def broadcast(
        cls,
        *,
        type: Literal["stdout", "stderr"],
        execution_id: str,
        stage_id: str,
        log: str,
    ):
        msg = serialize(
            dict(type=type, log=log, execution_id=execution_id, stage_id=stage_id)
        )
        for listener in cls.listeners:
            try:
                listener.send(msg)
            except Exception:
                pass

    def __init__(
        self,
        *,
        main_controller: MainController,
        sys_stdout_write,
        sys_stderr_write,
    ):
        self.execution_logs_repository = main_controller.execution_logs_repository
        self.execution_repository = main_controller.execution_repository
        self.sys_stdout_write = sys_stdout_write
        self.sys_stderr_write = sys_stderr_write

    def patched_stderr_write(self, raw: Union[str, bytearray]) -> int:
        return self._wrapper("stderr", self.sys_stderr_write, raw)

    def patched_stdout_write(self, raw: Union[str, bytearray]) -> int:
        return self._wrapper("stdout", self.sys_stdout_write, raw)

    def _wrapper(
        self,
        type: Literal["stdout", "stderr"],
        sys_write: Callable,
        raw: Union[str, bytearray],
    ):
        text = raw.decode("utf-8") if not isinstance(raw, str) else raw

        if IS_PRODUCTION:
            text = GLOBAL_MASKER.mask(text)

        try:
            short_id = self._capture_stdio(type, text).split(sep="-")[0]
            term_output = f"[RUN {short_id}] {text}" if text.strip() else text
            sys_write(term_output)
        except ExecutionNotFound:
            sys_write(text)
        except Exception as e:
            AbstraLogger.capture_exception(e)
        finally:
            sys.stdout.flush()
            return len(text)

    def _capture_stdio(self, type: Literal["stderr", "stdout"], text: str) -> str:
        execution = SDKContextStore.get_execution()
        self.execution_logs_repository.insert_stdio(execution.id, type, text)
        self.broadcast(
            type=type, log=text, execution_id=execution.id, stage_id=execution.stage_id
        )

        return execution.id



import base64
import json
from io import BytesIO
from typing import TYPE_CHECKING, Dict, List, Tuple, Union

from abstra_internals.email_signer import verify_email
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.utils.insensitive_dict import CaseInsensitiveDict

if TYPE_CHECKING:
    from abstra_internals.controllers.execution_client_hook import HookClient


class HookSDKController:
    def __init__(self, client: "HookClient") -> None:
        self.client = client

    def set_response(self, status: int, body: str, headers: Dict[str, str]):
        self.client.set_response(status, body, headers)

    def get_raw_request(self) -> Tuple[str, Dict[str, str], CaseInsensitiveDict]:
        request = self.client.get_request()
        return (
            request.body,
            request.query_params,
            CaseInsensitiveDict(request.headers),
        )

    def get_request(
        self,
    ) -> Tuple[Union[str, List, Dict], Dict[str, str], CaseInsensitiveDict]:
        if self.client.context.mock_execution.test_request:
            test_request = self.client.context.mock_execution.test_request
            body, query, headers = (
                test_request.body,
                test_request.query_params,
                CaseInsensitiveDict(test_request.headers),
            )
        else:
            body, query, headers = self.get_raw_request()
        try:
            if "application/json" in headers["Content-Type"]:
                return json.loads(body), query, headers
            elif "multipart/form-data" in headers["Content-Type"]:
                return self._multipart_form_parse(body, headers), query, headers
            else:
                return body, query, headers
        except Exception:
            return body, query, headers

    def _multipart_form_parse(self, body: str, headers: CaseInsensitiveDict):
        from multipart import (
            MultipartParser,
            MultipartPart,
            parse_options_header,
            to_bytes,
        )

        _, options = parse_options_header(headers["Content-Type"])
        boundary = options["boundary"].encode("utf-8")
        p = MultipartParser(BytesIO(to_bytes(body)), boundary)  # TODO: direct from fs
        return [
            {"name": i.name, "value": i.value}
            for i in p
            if isinstance(i, MultipartPart)
        ]

    def get_email_request(self):
        body, _, headers = self.get_raw_request()
        b64content: str = json.loads(body).get("content")
        if not b64content:
            raise Exception("Bad request")

        signed_hash = headers.get("x-abstra-notification-signed-hash")
        if not signed_hash and IS_PRODUCTION:
            raise Exception("Email verification failed")

        if signed_hash:
            if not verify_email(jwt_header=signed_hash, content=b64content):
                raise Exception("Email verification failed")

        from mailparser import parse_from_string

        return parse_from_string(base64.b64decode(b64content).decode("utf-8"))



from typing import Optional, Union

from abstra_internals.controllers.execution_client import (
    ExecutionClient,
    HeadlessClient,
)
from abstra_internals.controllers.execution_target import ExecutionTarget
from abstra_internals.entities.execution import Execution, PreExecution
from abstra_internals.entities.execution_context import (
    ClientContext,
    JobContext,
    ScriptContext,
)
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import Stage, StageWithFile


class NotStartedException(Exception):
    pass


class ExecutionController:
    def __init__(
        self,
        *,
        repositories: Repositories,
    ) -> None:
        self.repositories = repositories

    def submit(
        self,
        stage: Stage,
        context: Union[ScriptContext, JobContext],
    ):
        return self.repositories.producer.submit(
            PreExecution(
                context=context,
                stage_id=stage.id,
            )
        )

    def run(
        self,
        *,
        stage: StageWithFile,
        client: Optional[ExecutionClient] = None,
        context: Optional[ClientContext] = None,
    ):
        if not client:
            client = HeadlessClient()

        execution = Execution.create(
            context=context,
            stage_id=stage.id,
        )

        ExecutionTarget(
            stage=stage,
            client=client,
            execution=execution,
            repositories=self.repositories,
        )



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Tuple

from abstra_internals.interface.sdk import user_exceptions
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.repositories.tasks import TaskDTO


@dataclass
class DataRequestFilter:
    stage: Optional[List[str]]
    status: Optional[List[str]]
    start_date: Optional[str]
    end_date: Optional[str]

    @staticmethod
    def from_dict(data: dict) -> "DataRequestFilter":
        return DataRequestFilter(
            stage=data.get("stage", None),
            status=data.get("status", None),
            start_date=data.get("startDate", None),
            end_date=data.get("endDate", None),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class DataRequest:
    filter: DataRequestFilter
    limit: int
    offset: int

    @staticmethod
    def from_dict(data: dict) -> "DataRequest":
        return DataRequest(
            filter=DataRequestFilter.from_dict(data.get("filter", {})),
            limit=data.get("limit", 10),
            offset=data.get("offset", 0),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


class ListTasksItem(TaskDTO):
    target_stage_title: str
    target_stage_type: str
    source_stage_title: Optional[str]
    source_stage_type: Optional[str]


class TasksController:
    def __init__(
        self,
        repositories: Repositories,
    ) -> None:
        self.repos = repositories

    def get_stage(self, stage_id: str):
        project = ProjectRepository().load()
        stage = project.get_stage(stage_id)

        if not stage:
            raise Exception(f"Stage {stage_id} not found")

        return {
            "title": stage.title,
            "type_name": "tasklet" if stage.type_name == "script" else stage.type_name,
        }

    def get_nullable_stage(self, stage_id: Optional[str]):
        if not stage_id:
            return {
                "title": None,
                "type_name": None,
            }
        return self.get_stage(stage_id)

    def list_tasks(
        self, req: Optional[DataRequest] = None
    ) -> Tuple[List[ListTasksItem], int]:
        items = []
        for task in self.repos.tasks.get_all_tasks():
            if (
                req
                and req.filter.stage
                and task.target_stage_id not in req.filter.stage
            ):
                continue
            if req and req.filter.status and task.status not in req.filter.status:
                continue
            if req and (
                req.filter.start_date
                and datetime.fromisoformat(task.created.at)
                < datetime.fromisoformat(req.filter.start_date)
            ):
                continue
            if req and (
                req.filter.end_date
                and datetime.fromisoformat(task.created.at)
                > datetime.fromisoformat(req.filter.end_date)
            ):
                continue

            target_stage = self.get_stage(task.target_stage_id)
            source_stage = self.get_nullable_stage(task.created.by_stage_id)

            items.append(
                ListTasksItem(
                    **task.dump(),
                    target_stage_type=target_stage["type_name"],
                    target_stage_title=target_stage["title"],
                    source_stage_type=source_stage["type_name"],
                    source_stage_title=source_stage["title"],
                )
            )

        if not req:
            return items, len(items)

        return items[req.offset : req.offset + req.limit], len(items)

    def get_stage_tasks(self, stage_id) -> List[ListTasksItem]:
        target_stage = self.get_stage(stage_id)
        tasks = self.repos.tasks.get_stage_tasks(stage_id)

        if not target_stage["title"]:
            raise Exception(f"Stage {stage_id} not found")

        return [
            ListTasksItem(
                **task.dump(),
                target_stage_title=target_stage["title"],
                target_stage_type=target_stage["type_name"],
                source_stage_title=self.get_nullable_stage(task.created.by_stage_id)[
                    "title"
                ],
                source_stage_type=self.get_nullable_stage(task.created.by_stage_id)[
                    "type_name"
                ],
            )
            for task in tasks
        ]

    def get_sent_tasks(self, stage_id) -> List[ListTasksItem]:
        all_tasks, _ = self.list_tasks()
        tasks_with_executions = [
            (task, self.repos.execution.get(task.created.by_execution_id))
            for task in all_tasks
            if task.created.by_execution_id is not None
        ]

        return [
            ListTasksItem(
                **task.dump(),
                target_stage_title=task.target_stage_title,
                target_stage_type=task.target_stage_title,
                source_stage_title=self.get_nullable_stage(task.created.by_stage_id)[
                    "title"
                ],
                source_stage_type=self.get_nullable_stage(task.created.by_stage_id)[
                    "type_name"
                ],
            )
            for task, execution in tasks_with_executions
            if execution and execution.stage_id == stage_id
        ]

    def update_task_status(self, task_id: str, status: str) -> None:
        if status == "completed":
            self.repos.tasks.complete_task(task_id, None, None)
        elif status == "pending":
            self.repos.tasks.set_task_to_pending(task_id)
        else:
            raise user_exceptions.TaskInvalidStatus(status)

    def create_task(
        self,
        name: str,
        stage_id: str,
        payload: dict,
        source_stage_id: Optional[str] = None,
        execution_id: Optional[str] = None,
    ) -> TaskDTO:
        task = self.repos.tasks.send_task(
            name, payload, stage_id, source_stage_id, execution_id
        )
        return task
